from typing import Any, ClassVar
import warnings
import platform
import numpy as np

try:
    import mlx.core as mx
except ModuleNotFoundError as e:
    print(f"mlx library not installed. Try 'pip install mlx'.")


class MetalNotSupportedWarning(UserWarning):
    _has_warned: ClassVar[bool] = False

    def __init__(self, message=None):
        system = platform.system()
        machine = platform.machine()
        arch = platform.processor() or machine

        default_message = (
            f"Metal GPU acceleration is not supported on this system "
            f"({system}, {arch}). Falling back to CPU, which may be slower "
            f"than native CPU due to lazy execution overhead."
        )
        super().__init__(message or default_message)


def check_metal_availability() -> None:
    if MetalNotSupportedWarning._has_warned:
        return
    if not mx.metal.is_available():
        MetalNotSupportedWarning._has_warned = True
        warnings.warn(MetalNotSupportedWarning(), stacklevel=2)


def is_cpu_op(*tensor_or_any) -> bool:
    for t in tensor_or_any:
        device = getattr(t, "device", None)
        if device is None:
            if isinstance(t, mx.array):
                return False
        else:
            if device == "gpu":
                return False
    return True


def is_gpu_op(*tensor_or_any) -> bool:
    for t in tensor_or_any:
        device = getattr(t, "device", None)
        if device is None:
            if isinstance(t, mx.array):
                return True
        else:
            if device == "gpu":
                return True
    return False


def parse_mlx_indexing(index: Any) -> Any:
    if isinstance(index, np.ndarray):
        raise TypeError(
            "GPU tensors do not support CPU tensor or NumPy array indexing. "
            + "Convert to GPU tensors."
        )

    if isinstance(index, tuple):
        parsed = []
        for i, idx in enumerate(index):
            if isinstance(idx, np.ndarray):
                raise ValueError(f"NumPy array indexing found at {i}-th index.")

            if isinstance(idx, bool):
                parsed.append(1 if idx else 0)

            elif isinstance(idx, mx.array) and idx.dtype == mx.bool_:
                parsed.append(mx.array(np.flatnonzero(idx.tolist()), dtype=mx.int32))

            elif isinstance(idx, list) and all(isinstance(i, bool) for i in idx):
                mask = mx.array(idx, dtype=mx.bool_)
                parsed.append(mx.array(np.flatnonzero(mask.tolist()), dtype=mx.int32))

            elif isinstance(idx, list):
                parsed.append(mx.array(idx, dtype=mx.int32))

            else:
                parsed.append(idx)

        return tuple(parsed)

    elif isinstance(index, bool):
        return 1 if index else 0

    elif isinstance(index, mx.array) and index.dtype == mx.bool_:
        return mx.array(np.flatnonzero(index.tolist()), dtype=mx.int32)

    elif isinstance(index, list) and all(isinstance(i, bool) for i in index):
        mask = mx.array(index, dtype=mx.bool_)
        return mx.array(np.flatnonzero(mask.tolist()), dtype=mx.int32)

    elif isinstance(index, list):
        return mx.array(index, dtype=mx.int32)

    return index



from typing import Callable, List, Any

import lucid
import lucid.nn as nn

from lucid._tensor import Tensor
from lucid.types import _BuiltinNumeric, Numeric


class Compose:
    def __init__(self, transforms: List[Callable[[Any], Any]]) -> None:
        for t in transforms:
            if not isinstance(t, nn.Module):
                raise TypeError(
                    f"Expected a Module instance, got {type(t).__name__}",
                )
        self.transforms = transforms

    def __call__(self, x: Any) -> Any | Tensor:
        for transform in self.transforms:
            x = transform(x)
        return x

    def __repr__(self) -> str:
        transform_names = ", ".join(
            [type(transform).__name__ for transform in self.transforms]
        )
        return f"{type(self).__name__}([{transform_names}])"


class ToTensor(nn.Module):
    def __init__(
        self,
        requires_grad: bool = False,
        keep_grad: bool = False,
        dtype: _BuiltinNumeric | Numeric | None = None,
    ) -> None:
        super().__init__()
        self.requires_grad = requires_grad
        self.keep_grad = keep_grad
        self.dtype = dtype

    def forward(self, x: Any) -> Tensor:
        if self.dtype is None:
            if hasattr(x, "dtype"):
                self.dtype = x.dtype
            else:
                self.dtype = None

        return lucid.to_tensor(x, self.requires_grad, self.keep_grad, self.dtype)



from .linear import *
from .conv import *
from .activation import *
from .pool import *
from .norm import *
from .drop import *
from .loss import *
from .vision import *
from .attention import *
from .transformer import *
from .sparse import *
from .einops import *



from typing import Never
import lucid
from lucid._tensor import Tensor


def _prob_check(p: float) -> Never:
    if not 0 <= p < 1:
        raise ValueError("Dropout probability `p` must be in the range [0, 1).")


def dropout(input_: Tensor, p: float = 0.5, training: bool = True) -> Tensor:
    _prob_check(p)
    if not training:
        return input_

    mask = (lucid.random.rand(*input_.shape) > p).free()
    scale = 1.0 / (1 - p)
    return input_ * mask * scale


def dropoutnd(input_: Tensor, p: float = 0.5, training: bool = True) -> Tensor:
    _prob_check(p)
    if not training:
        return input_

    spatial_dim = input_.ndim - 2
    mask = (lucid.random.rand(*input_.shape[:2], *(1,) * spatial_dim) > p).free()
    scale = 1.0 / (1 - p)
    return input_ * mask * scale


def alpha_dropout(input_: Tensor, p: float = 0.5, training: bool = True) -> Tensor:
    _prob_check(p)
    if not training:
        return input_

    _alpha = -1.7580993408473766
    _lambda = 1.0507009873554805

    mask = (lucid.random.rand(*input_.shape) > p).free()
    scale = 1.0 / (1 - p)

    dropped = input_ * mask * scale
    noise = (1 - mask) * _alpha * _lambda
    return dropped + noise


def drop_block(
    input_: Tensor, block_size: int, p: float = 0.1, eps: float = 1e-7
) -> Tensor:
    _, _, h, w = input_.shape
    gamma = (
        p
        * (h * w)
        / (block_size**2)
        / ((h - block_size + 1) * (w - block_size + 1) + eps)
    )
    mask = (lucid.random.rand(*input_.shape) < gamma).astype(float)

    pad = block_size // 2
    padded_mask = lucid.pad(mask, ((0, 0), (0, 0), (pad, pad), (pad, pad)))
    block_mask = lucid.zeros_like(mask)

    for i in range(block_size):
        for j in range(block_size):
            block_mask = lucid.maximum(
                block_mask, padded_mask[:, :, i : i + h, j : j + w]
            )

    block_mask = (1 - block_mask).free()
    return input_ * block_mask / (1 - p)


def drop_path(input_: Tensor, p: float = 0.0, scale_by_keep: bool = True) -> Tensor:
    if p == 0.0:
        return input_

    keep_prob = 1 - p
    shape = (input_.shape[0],) + (1,) * (input_.ndim - 1)

    random_tensor = lucid.random.bernoulli(keep_prob * lucid.ones(shape)).free()
    if keep_prob > 0.0 and scale_by_keep:
        random_tensor /= keep_prob

    return input_ * random_tensor



from lucid._tensor import Tensor
from lucid.linalg import _func


# fmt: off
__all__ = [
    "inv", "det", "solve", "cholesky", "norm", "qr", "svd", "matrix_power", "pinv"
]
# fmt: on


def inv(a: Tensor, /) -> Tensor:
    return _func.inv()(a)


def det(a: Tensor, /) -> Tensor:
    return _func.det()(a)


def solve(a: Tensor, b: Tensor, /) -> Tensor:
    return _func.solve()(a, b)


def cholesky(a: Tensor, /) -> Tensor:
    return _func.cholesky()(a)


def norm(
    a: Tensor,
    /,
    ord: int = 2,
    axis: tuple[int, ...] | int | None = None,
    keepdims: bool = False,
) -> Tensor:
    return _func.norm(ord, axis, keepdims)(a)


def eig(a: Tensor, /, eps: float = 1e-12) -> tuple[Tensor, Tensor]:
    return _func.eig(eps)(a)


def qr(a: Tensor, /) -> tuple[Tensor, Tensor]:
    return _func.qr()(a)


def svd(a: Tensor, /, full_matrices: bool = True) -> tuple[Tensor, Tensor, Tensor]:
    return _func.svd(full_matrices)(a)


def matrix_power(a: Tensor, /, n: int) -> Tensor:
    return _func.matrix_power(n)(a)


def pinv(a: Tensor, /, rcond: float = 1e-12) -> Tensor:
    return _func.pinv(rcond)(a)



from lucid.optim.base import Optimizer

from lucid.optim._sgd import *
from lucid.optim._prop import *
from lucid.optim._adam import *
from lucid.optim._ada import *

from lucid.optim import lr_scheduler



from types import ModuleType
from functools import partial
import numpy as np

from lucid.types import _NumPyArray, _MLXArray
from lucid._tensor import Tensor

from lucid._backend.core import (
    operation,
    fallback,
    func_op,
    binary_func_op,
    unary_func_op,
    _GradFuncType,
    _FuncOpReturnType,
)
from lucid._backend.metal import mx


class inv(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.linalg.inv(a.data))
        return self.result, self.__grad_cpu__

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.linalg.inv(a.data))
        return self.result, self.__grad_gpu__

    def __grad_cpu__(self) -> _GradFuncType:
        return -np.dot(np.dot(self.result.data.T, self.result.grad), self.result.data)

    def __grad_gpu__(self) -> _GradFuncType:
        return -mx.matmul(
            mx.matmul(self.result.data.T, self.result.grad), self.result.data
        )

    def __flops__(self, a: Tensor) -> int:
        return int((2 / 3) * a.shape[-1] ** 3)


class det(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.linalg.det(a.data))
        return self.result, partial(self.__grad_cpu__, a=a)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        _, _, U = mx.linalg.lu(a.data)
        diag = mx.diagonal(U)

        self.result = Tensor(mx.prod(diag))
        return self.result, partial(self.__grad_gpu__, a=a)

    def __grad_cpu__(self, a: Tensor) -> _GradFuncType:
        grad = self.result.grad
        invA_T = np.transpose(np.linalg.inv(a.data))
        return grad * invA_T

    def __grad_gpu__(self, a: Tensor) -> _GradFuncType:
        grad = self.result.grad
        invA_T = mx.transpose(mx.linalg.inv(a.data))
        return grad * invA_T

    def __flops__(self, a: Tensor) -> int:
        return int((1 / 3) * a.shape[-1] ** 3)


class solve(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op()
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.linalg.solve(a.data, b.data))
        return self.result, partial(self.__grad_cpu__, a=a)

    @binary_func_op(device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.linalg.solve(a.data, b.data))
        return self.result, partial(self.__grad_gpu__, a=a)

    def __grad_cpu__(self, a: Tensor) -> _GradFuncType:
        grad = self.result.grad
        x = self.result.data
        inv_a = np.linalg.inv(a.data)

        a_grad = -inv_a @ (grad @ x.T) @ inv_a
        b_grad = inv_a @ grad

        return a_grad, b_grad

    def __grad_gpu__(self, a: Tensor) -> _GradFuncType:
        grad = self.result.grad
        x = self.result.data
        inv_a = mx.linalg.inv(a.data)

        a_grad = -mx.matmul(inv_a, mx.matmul(mx.matmul(grad, x.T), inv_a))
        b_grad = mx.matmul(inv_a, grad)

        return a_grad, b_grad

    def __flops__(self, a: Tensor, b: Tensor) -> int:
        n = a.shape[-1]
        m = b.shape[-1] if b.ndim >= 2 else 1
        return int((2 / 3) * n**3 + 2 * n**2 * m)


class cholesky(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.linalg.cholesky(a.data))
        return self.result, partial(self.__grad__, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.linalg.cholesky(a.data))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        L = self.result.data
        grad_L = self.result.grad

        L_inv = lib_.linalg.inv(L)
        inner = L.T @ grad_L
        sym = 0.5 * (inner + inner.T)

        return L_inv.T @ (sym @ L_inv)

    def __flops__(self, a: Tensor) -> int:
        return int((1 / 3) * a.shape[-1] ** 3)


@fallback
class norm(operation):
    def __init__(
        self,
        ord: int = 2,
        axis: int | tuple[int, ...] | None = None,
        keepdims: bool = False,
    ) -> None:
        super().__init__()
        self.ord = ord
        self.axis = axis
        self.keepdims = keepdims

        if not isinstance(self.ord, int):
            raise NotImplementedError("Only integer p-norms are supported.")

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        result_data = np.linalg.norm(
            a.data, ord=self.ord, axis=self.axis, keepdims=self.keepdims
        )
        self.result = Tensor(result_data)
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        fallback_ = self.axis is None or (
            isinstance(self.axis, (tuple, list)) and a.ndim > 1
        )
        if fallback_:
            result_data = np.linalg.norm(
                a.data, ord=self.ord, axis=self.axis, keepdims=self.keepdims
            )
            self.result = Tensor(result_data, device="gpu")
        else:
            result_data = mx.linalg.norm(
                a.data, ord=self.ord, axis=self.axis, keepdims=self.keepdims
            )
            self.result = Tensor(result_data)

        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        x = a.data
        r = self.result.data
        grad_output = self.result.grad

        ord = self.ord
        axis = self.axis
        keepdims = self.keepdims

        if ord == 2:
            denom = r
            if not keepdims and axis is not None:
                denom = lib_.expand_dims(r, axis=axis)
            grad = lib_.where(lib_.all(r != 0), x / denom, lib_.zeros_like(x))

        elif ord == 1:
            grad = lib_.sign(x)

        else:
            denom = r
            if not keepdims and axis is not None:
                denom = lib_.expand_dims(r, axis=axis)
            grad = lib_.where(
                lib_.all(r != 0),
                (lib_.abs(x) ** (ord - 1)) * lib_.sign(x) / (denom ** (ord - 1)),
                lib_.zeros_like(x),
            )

        if axis is not None and not keepdims:
            grad_output = lib_.expand_dims(grad_output, axis=axis)

        return grad * grad_output

    def __flops__(self, a: Tensor) -> int:
        if self.axis is None:
            reduced_size = a.size
        else:
            if isinstance(self.axis, int):
                self.axis = (self.axis,)
            reduced_size = 1
            for ax in self.axis:
                reduced_size *= a.shape[ax]

        out_size = a.size // reduced_size
        if self.ord == 1:
            flops_per_out = 2 * reduced_size - 1
        elif self.ord == 2:
            flops_per_out = 2 * reduced_size
        else:
            flops_per_out = 3 * reduced_size

        return out_size * flops_per_out


@fallback
class eig(operation):
    def __init__(self, eps: float) -> None:
        super().__init__()
        self.eps = eps

    def _unified(self, a: Tensor) -> tuple[_NumPyArray, _NumPyArray]:
        self.ndim = a.shape[-2]
        eigvals, eigvecs = np.linalg.eig(a.data)
        eigvecs = eigvecs / np.linalg.norm(eigvecs, axis=-2, keepdims=True)

        return eigvals, eigvecs

    @func_op(n_in=1, n_ret=2)
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self._eigvals, self._eigvecs = self._unified(a)
        self.result = (Tensor(self._eigvals), Tensor(self._eigvecs))

        return (
            (self.result[0], self.__grad_eigvals__),
            (self.result[1], self.__grad_eigvecs__),
        )

    @func_op(n_in=1, n_ret=2, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self._eigvals, self._eigvecs = self._unified(a)
        self.result = (
            Tensor(self._eigvals, device="gpu"),
            Tensor(self._eigvecs, device="gpu"),
        )

        return (
            (self.result[0], partial(self.__grad_eigvals__, _fallback=True)),
            (self.result[1], partial(self.__grad_eigvecs__, _fallback=True)),
        )

    def __grad_eigvals__(self, _fallback: bool = False) -> _GradFuncType:
        eigvals = self.result[0]
        grad = np.einsum(
            "...k,...ki,...kj->...ij",
            np.array(eigvals.grad),
            self._eigvecs,
            self._eigvecs,
        )
        return grad if not _fallback else mx.array(grad)

    def __grad_eigvecs__(self, _fallback: bool = False) -> _GradFuncType:
        eigvals, eigvecs = self._eigvals, self._eigvecs
        eigvecs_t = self.result[1]

        eigval_diffs = eigvals[..., :, np.newaxis] - eigvals[..., np.newaxis, :]
        eigval_diffs += np.eye(self.ndim)[..., :, :] * self.eps

        inv_eigval_diffs = 1.0 / eigval_diffs
        for index in np.ndindex(inv_eigval_diffs.shape[:-2]):
            np.fill_diagonal(inv_eigval_diffs[index], 0.0)

        outer_prods = np.einsum("...ip,...jq->...pqij", eigvecs, eigvecs)
        S = np.einsum("...kp,...pqij->...pij", inv_eigval_diffs, outer_prods)

        grad = np.einsum(
            "...pk,...pij,...ki->...ij", np.array(eigvecs_t.grad), S, eigvecs
        )
        return grad if not _fallback else mx.array(grad)

    def __flops__(self, a: Tensor) -> int:
        return 9 * a.shape[-1] ** 3


class qr(operation):
    def __init__(self) -> None:
        super().__init__()

    @func_op(n_in=1, n_ret=2)
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.Q, self.R = np.linalg.qr(a.data)
        self.result = (Tensor(self.Q), Tensor(self.R))

        return (
            (self.result[0], partial(self.__grad_q__, lib_=np)),
            (self.result[1], partial(self.__grad_r__, lib_=np)),
        )

    @func_op(n_in=1, n_ret=2, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.Q, self.R = mx.linalg.qr(a.data)
        self.result = (Tensor(self.Q), Tensor(self.R))

        return (
            (self.result[0], partial(self.__grad_q__, lib_=mx)),
            (self.result[1], partial(self.__grad_r__, lib_=mx)),
        )

    def __grad_q__(self, lib_: ModuleType) -> _GradFuncType:
        grad_q = self.result[0].grad
        qt_grad_q = lib_.einsum("...ik,...kj->...ij", self.Q.swapaxes(-1, -2), grad_q)
        qt_grad_q_r = lib_.einsum("...ij,...jk->...ik", qt_grad_q, self.R)

        return lib_.einsum("...ij,...jk->...ik", grad_q, self.R) - lib_.einsum(
            "...ij,...jk->...ik", self.Q, qt_grad_q_r
        )

    def __grad_r__(self, lib_: ModuleType) -> _GradFuncType:
        grad_r = self.result[1].grad
        return lib_.einsum("...ij,...jk->...ik", self.Q, grad_r)

    def __flops__(self, a: Tensor) -> int:
        m, n = a.shape[-2:]
        return int(2 * n**2 * (m - n / 3))


@fallback
class svd(operation):
    def __init__(self, full_matrices: bool) -> None:
        super().__init__()
        self.full_matrices = full_matrices

    @func_op(n_in=1, n_ret=3)
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.U, self.S, self.VT = np.linalg.svd(a.data, self.full_matrices)
        self.result = (Tensor(self.U), Tensor(self.S), Tensor(self.VT))

        return (
            (self.result[0], partial(self.__grad_u__, lib_=np)),
            (self.result[1], partial(self.__grad_s__, lib_=np)),
            (self.result[2], partial(self.__grad_vt__, lib_=np)),
        )

    @func_op(n_in=1, n_ret=3, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        if self.full_matrices:
            U, S, VT = np.linalg.svd(np.array(a.data), full_matrices=True)
            self.U, self.S, self.VT = mx.array(U), mx.array(S), mx.array(VT)
        else:
            self.U, self.S, self.VT = mx.linalg.svd(a.data)

        self.result = (Tensor(self.U), Tensor(self.S), Tensor(self.VT))

        return (
            (self.result[0], partial(self.__grad_u__, lib_=mx)),
            (self.result[1], partial(self.__grad_s__, lib_=mx)),
            (self.result[2], partial(self.__grad_vt__, lib_=mx)),
        )

    def __grad_u__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.einsum(
            "...ik,...k,...jk->...ij",
            self.result[0].grad,
            self.S,
            self.VT.swapaxes(-1, -2),
        )

    def __grad_s__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.einsum(
            "...ik,...k,...jk->...ij",
            self.U,
            self.result[1].grad,
            self.VT.swapaxes(-1, -2),
        )

    def __grad_vt__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.einsum(
            "...ik,...k,...jk->...ij",
            self.U,
            self.S,
            self.result[2].grad.swapaxes(-1, -2),
        )

    def __flops__(self, a: Tensor) -> int:
        m, n = a.shape[-2:]
        if m < n:
            m, n = n, m
        return int(4 * m**2 * n + 8 * m * n**2 + 9 * n**3)


class matrix_power(operation):
    def __init__(self, n: int) -> None:
        super().__init__()
        self.n = n

    def _gpu_matrix_pow(self, arr: _MLXArray, n: int) -> _MLXArray:
        if n == 0:
            return mx.eye(arr.shape[0], dtype=arr.dtype)
        elif n < 0:
            arr = mx.linalg.inv(arr)
            n = -n

        result = arr
        for _ in range(n - 1):
            result = result @ arr
        return result

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.linalg.matrix_power(a.data, n=self.n))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(self._gpu_matrix_pow(a.data, n=self.n))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        grad = lib_.zeros_like(a.data)
        if self.n == 0:
            return grad
        else:
            for k in range(abs(self.n)):
                left_exp = self.n - lib_.sign(self.n) * k - lib_.sign(self.n)
                right_exp = lib_.sign(self.n) * k

                left = (
                    np.linalg.matrix_power(a.data, left_exp)
                    if lib_ is np
                    else self._gpu_matrix_pow(a.data, left_exp.item())
                )
                right = (
                    np.linalg.matrix_power(a.data, right_exp)
                    if lib_ is np
                    else self._gpu_matrix_pow(a.data, right_exp.item())
                )

                grad += left @ self.result.grad @ right
            if self.n < 0:
                grad = -grad

        return grad

    def __flops__(self, a: Tensor) -> int:
        d = a.shape[-1]
        num_mults = max(0, self.n.bit_length() - 1)
        return 2 * num_mults * d**3


class pinv(operation):
    def __init__(self, rcond: float) -> None:
        super().__init__()
        self.rcond = rcond

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.linalg.pinv(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        U, S, VT = mx.linalg.svd(a.data)
        S_inv = mx.where(S > self.rcond, 1.0 / S, 0.0)
        S_inv_mat = mx.diag(S_inv)

        self.result = Tensor(VT.T @ S_inv_mat @ U.T)
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        U, S, Vh = (
            np.linalg.svd(a.data, full_matrices=False)
            if lib_ is np
            else mx.linalg.svd(a.data)
        )
        S_inv_squared = lib_.diag(1 / (S**2))
        term_1 = (
            Vh.T
            @ S_inv_squared
            @ U.T
            @ self.result.grad.T
            @ (lib_.eye(a.shape[0]) - a.data @ self.result.data)
        )
        term_2 = (
            (lib_.eye(a.shape[1]) - self.result.data @ a.data)
            @ self.result.grad.T
            @ U
            @ S_inv_squared
            @ Vh
        )
        grad = (
            -self.result.data.T @ self.result.grad.T @ self.result.data.T
            + term_1
            + term_2
        ).T
        return grad

    def __flops__(self, a: Tensor) -> int:
        m, n = a.shape[-2:]
        if m < n:
            m, n = n, m
        return int(20 * m * n**2 + 20 * n**3)



import lucid

from lucid._tensor import Tensor


def relu(input_: Tensor) -> Tensor:
    return lucid.maximum(0, input_)


def leaky_relu(input_: Tensor, negative_slope: float = 0.01) -> Tensor:
    mask = input_ > 0
    out = input_ * mask + input_ * negative_slope * (1 - mask)
    return out


def elu(input_: Tensor, alpha: float = 1.0) -> Tensor:
    mask = input_ >= 0
    pos = input_ * mask
    neg = alpha * (lucid.exp(input_) - 1) * (1 - mask)
    return pos + neg


def selu(input_: Tensor) -> Tensor:
    _scale = 1.0507009873554805
    _alpha = 1.6732632423543772

    mask = input_ >= 0
    pos = _scale * input_ * mask
    neg = _scale * _alpha * (lucid.exp(input_) - 1) * (1 - mask)
    return pos + neg


def gelu(input_: Tensor) -> Tensor:
    c = lucid.sqrt(2 / lucid.pi).free()
    return 0.5 * input_ * (1 + lucid.tanh(c * (input_ + 0.044715 * input_**3)))


def sigmoid(input_: Tensor) -> Tensor:
    return 1 / (1 + lucid.exp(-input_))


def tanh(input_: Tensor) -> Tensor:
    return lucid.tanh(input_)


def softmax(input_: Tensor, axis: int = -1) -> Tensor:
    input_max = lucid.max(input_, axis=axis, keepdims=True)
    input_stable = input_ - input_max

    e_input = lucid.exp(input_stable)
    sum_e_input = e_input.sum(axis=axis, keepdims=True)

    output = e_input / sum_e_input
    return output



import math
from functools import partial
from typing import Callable, ClassVar, Literal, NamedTuple

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid import register_model
from lucid._tensor import Tensor


__all__ = ["CSPNet", "csp_resnet_50", "csp_resnext_50_32x4d", "csp_darknet_53"]


class _ConvBNAct(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        k: int = 3,
        s: int = 1,
        p: int | None = None,
        groups: int = 1,
        bias: bool = False,
        norm: type[nn.Module] = nn.BatchNorm2d,
        act: type[nn.Module] = nn.ReLU,
    ) -> None:
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=k,
            stride=s,
            padding=p if p is not None else "same",
            groups=groups,
            bias=bias,
        )
        self.norm = norm(out_channels)
        self.act = act()

    def forward(self, x: Tensor) -> Tensor:
        return self.act(self.norm(self.conv(x)))


class _Bottleneck(nn.Module):
    expansion: ClassVar[int] = 4

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int = 1,
        downsample: nn.Module | None = None,
        norm: type[nn.Module] = nn.BatchNorm2d,
        act: type[nn.Module] = nn.ReLU,
        groups: int = 1,
        base_width: int = 64,
        **kwargs,
    ) -> None:
        super().__init__()
        width = int(out_channels * (base_width / 64.0)) * groups

        self.conv1 = _ConvBNAct(in_channels, width, k=1, s=1, p=0, norm=norm, act=act)
        self.conv2 = _ConvBNAct(
            width, width, k=3, s=stride, p=1, groups=groups, norm=norm, act=act
        )
        self.conv3 = nn.Conv2d(
            width, out_channels * self.expansion, kernel_size=1, bias=False
        )
        self.norm3 = norm(out_channels * self.expansion)
        self.act = act()
        self.downsample = downsample

    def forward(self, x: Tensor) -> Tensor:
        identity = x
        out = self.conv1(x)
        out = self.conv2(out)
        out = self.norm3(self.conv3(out))

        if self.downsample is not None:
            identity = self.downsample(x)
        out += identity
        return self.act(out)


class _DarknetBottleneck(nn.Module):
    expansion: ClassVar[int] = 1

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int = 1,
        norm: type[nn.Module] = nn.BatchNorm2d,
        act: type[nn.Module] = partial(nn.LeakyReLU, negative_slope=0.1),
        **kwargs,
    ) -> None:
        super().__init__()
        self.conv1 = _ConvBNAct(
            in_channels, out_channels, k=1, s=1, p=0, norm=norm, act=act
        )
        self.conv2 = _ConvBNAct(
            out_channels, out_channels, k=3, s=stride, p=1, norm=norm, act=act
        )

        self.use_proj = not (stride == 1 and in_channels == out_channels)
        self.downsample = (
            nn.Identity()
            if not self.use_proj
            else nn.Sequential(
                nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False),
                norm(out_channels),
            )
        )
        self.act = act()

    def forward(self, x: Tensor) -> Tensor:
        identity = x
        out = self.conv1(x)
        out = self.conv2(out)
        if self.use_proj:
            identity = self.downsample(x)

        out += identity
        return self.act(out)


class _StackOut(NamedTuple):
    module: nn.Module
    out_channels: int


def _resnet_stack_factory(
    block_cls: type[nn.Module],
    norm: type[nn.Module] = nn.BatchNorm2d,
    act: type[nn.Module] = nn.ReLU,
    groups: int = 1,
    base_width: int = 64,
    **kwargs,
) -> Callable[[int, int, int], _StackOut]:
    expansion = getattr(block_cls, "expansion", 1)

    def make_stack(
        in_channels: int, num_layers: int, stride_first: int = 1
    ) -> _StackOut:
        channels = math.ceil(in_channels / expansion)
        layers = []

        down = None
        if in_channels != channels * expansion:
            down = nn.Sequential(
                nn.Conv2d(in_channels, channels * expansion, kernel_size=1, bias=False),
                norm(channels * expansion),
            )

        base_kwargs = dict(norm=norm, act=act, groups=groups, base_width=base_width)
        layers.append(
            block_cls(
                in_channels,
                channels,
                stride=stride_first,
                downsample=down,
                **base_kwargs,
            )
        )
        in_channels = channels * expansion
        for _ in range(1, num_layers):
            layers.append(
                block_cls(
                    in_channels, channels, stride=1, downsample=None, **base_kwargs
                )
            )

        return _StackOut(nn.Sequential(*layers), in_channels)

    make_stack.required_multiple = expansion
    return make_stack


def _darknet_stack_factory(
    block_cls: type[nn.Module],
    norm: type[nn.Module] = nn.BatchNorm2d,
    act: type[nn.Module] = partial(nn.LeakyReLU, negative_slope=0.1),
    **kwargs,
) -> Callable[[int, int, int], _StackOut]:
    expansion = getattr(block_cls, "expansion", 1)

    def make_stack(
        in_channels: int, num_layers: int, stride_first: int = 1
    ) -> _StackOut:
        channels = in_channels // expansion
        base_kwargs = dict(downsample=None, norm=norm, act=act)

        layers = []
        layers.append(
            block_cls(
                in_channels, channels * expansion, stride=stride_first, **base_kwargs
            )
        )
        in_ch = channels * expansion
        for _ in range(1, num_layers):
            layers.append(
                block_cls(in_ch, channels * expansion, stride=1, **base_kwargs)
            )

        return _StackOut(nn.Sequential(*layers), in_ch)

    make_stack.required_multiple = expansion
    return make_stack


class _CSPStage(nn.Module):
    def __init__(
        self,
        in_channels: int,
        stage_width: int,
        num_layers: int,
        block_stack_fn: Callable[[int, int, int], _StackOut],
        split_ratio: float = 0.5,
        downsample: bool = False,
        norm: type[nn.Module] = nn.BatchNorm2d,
        act: type[nn.Module] = nn.ReLU,
        pre_kernel_size: int = 1,
    ) -> None:
        super().__init__()
        self.pre = _ConvBNAct(
            in_channels,
            stage_width,
            k=pre_kernel_size,
            s=(2 if downsample else 1),
            p=(pre_kernel_size - 1) // 2,
            norm=norm,
            act=act,
        )

        c1 = int(round(stage_width * split_ratio))
        c2 = stage_width - c1
        assert c1 > 0 and c2 > 0

        req = getattr(block_stack_fn, "required_multiple", 1)
        if c2 % req != 0:
            c2 = max(req, (c2 // req) * req)
            c1 = stage_width - c2
        self.c1, self.c2 = c1, c2

        self.part1_proj = _ConvBNAct(stage_width, c1, k=1, s=1, p=0, norm=norm, act=act)
        self.part2_proj = _ConvBNAct(stage_width, c2, k=1, s=1, p=0, norm=norm, act=act)

        stack_out = block_stack_fn(c2, num_layers, stride_first=1)
        self.block_stack = stack_out.module
        self.block_out = stack_out.out_channels

        self.merge = _ConvBNAct(
            c1 + self.block_out, stage_width, k=1, s=1, p=0, norm=norm, act=act
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.pre(x)
        y1 = self.part1_proj(x)
        y2 = self.part2_proj(x)
        y2 = self.block_stack(y2)

        y = lucid.concatenate([y1, y2], axis=1)
        y = self.merge(y)
        return y


class CSPNet(nn.Module):
    def __init__(
        self,
        in_channels: int = 3,
        stem_channels: int = 64,
        stage_defs: list[tuple[int, int, Callable, bool]] | None = None,
        num_classes: int = 1000,
        norm: type[nn.Module] = nn.BatchNorm2d,
        act: type[nn.Module] = nn.ReLU,
        split_ratio: float = 0.5,
        global_pool: Literal["avg", "max"] = "avg",
        dropout: float = 0.0,
        feature_channels: int | None = None,
        pre_kernel_size: int = 1,
    ) -> None:
        super().__init__()
        if stage_defs is None:
            stage_defs = []

        self.stem = nn.Sequential(
            _ConvBNAct(in_channels, stem_channels, k=3, s=2, norm=norm, act=act),
            _ConvBNAct(stem_channels, stem_channels, k=3, s=1, norm=norm, act=act),
        )

        stages = []
        in_ch = stem_channels
        for stage_width, num_layers, stack_fn, downsample in stage_defs:
            stages.append(
                _CSPStage(
                    in_ch,
                    stage_width,
                    num_layers,
                    stack_fn,
                    split_ratio=split_ratio,
                    downsample=downsample,
                    norm=norm,
                    act=act,
                    pre_kernel_size=pre_kernel_size,
                )
            )
            in_ch = stage_width
        self.stages = nn.Sequential(*stages)

        if feature_channels is not None and feature_channels != in_ch:
            self.pre_head = _ConvBNAct(
                in_ch, feature_channels, k=1, s=1, p=0, norm=norm, act=act
            )
            in_ch = feature_channels
        else:
            self.pre_head = nn.Identity()

        self.num_classes = num_classes
        self.head_pool = (
            nn.AdaptiveAvgPool2d((1, 1))
            if global_pool == "avg"
            else nn.AdaptiveMaxPool2d((1, 1))
        )
        self.head = nn.Sequential(
            nn.Flatten(),
            nn.Dropout(dropout) if dropout > 0.0 else nn.Identity(),
            nn.Linear(in_ch, num_classes),
        )

        self.apply(self.init_weights)

    def init_weights(self, m: nn.Module):
        if isinstance(m, nn.Conv2d):
            nn.init.kaiming_normal(m.weight, mode="fan_out")
        elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
            nn.init.constant(m.weight, 1.0)
            nn.init.constant(m.bias, 0.0)

    def forward_features(
        self, x: Tensor, return_stage_out: bool = False
    ) -> Tensor | list[Tensor]:
        feats = []
        x = self.stem(x)
        for stage in self.stages:
            x = stage(x)
            if return_stage_out:
                feats.append(x)

        return feats if return_stage_out else x

    def forward(self, x: Tensor) -> Tensor:
        x = self.forward_features(x, return_stage_out=False)
        x = self.pre_head(x)
        x = self.head_pool(x)
        x = self.head(x)
        return x


@register_model
def csp_resnet_50(
    num_classes: int = 1000, split_ratio: float = 0.5, stem_channels: int = 64, **kwargs
) -> CSPNet:
    res_stack = _resnet_stack_factory(_Bottleneck, groups=1, base_width=64, **kwargs)
    stage_defs = [
        (256, 3, res_stack, False),
        (512, 4, res_stack, True),
        (1024, 6, res_stack, True),
        (2048, 3, res_stack, True),
    ]
    return CSPNet(
        stem_channels=stem_channels,
        stage_defs=stage_defs,
        num_classes=num_classes,
        split_ratio=split_ratio,
        feature_channels=1024,
        **kwargs,
    )


@register_model
def csp_resnext_50_32x4d(
    num_classes: int = 1000, split_ratio: float = 0.5, stem_channels: int = 64, **kwargs
) -> CSPNet:
    resnext_stack = _resnet_stack_factory(
        _Bottleneck, groups=32, base_width=4, **kwargs
    )
    stage_defs = [
        (256, 3, resnext_stack, False),
        (512, 4, resnext_stack, True),
        (1024, 6, resnext_stack, True),
        (2048, 3, resnext_stack, True),
    ]
    return CSPNet(
        stem_channels=stem_channels,
        stage_defs=stage_defs,
        num_classes=num_classes,
        split_ratio=split_ratio,
        feature_channels=1024,
        **kwargs,
    )


@register_model
def csp_darknet_53(
    num_classes: int = 1000, split_ratio: float = 0.5, stem_channels: int = 32, **kwargs
) -> CSPNet:
    dark_stack = _darknet_stack_factory(_DarknetBottleneck, **kwargs)
    stage_defs = [
        (64, 1, dark_stack, True),
        (128, 2, dark_stack, True),
        (256, 8, dark_stack, True),
        (512, 8, dark_stack, True),
        (1024, 4, dark_stack, True),
    ]
    return CSPNet(
        stem_channels=stem_channels,
        stage_defs=stage_defs,
        num_classes=num_classes,
        split_ratio=split_ratio,
        feature_channels=1024,
        pre_kernel_size=3,
        **kwargs,
    )



from functools import reduce

import lucid

from lucid._tensor import Tensor
from lucid.types import _Scalar


def _calculate_fan_in_and_fan_out(tensor: Tensor) -> tuple[int, int]:
    if tensor.ndim == 2:
        fan_in = tensor.shape[0]
        fan_out = tensor.shape[1]

    elif tensor.ndim in {3, 4, 5}:
        kernel_prod = reduce(lambda x, y: x * y, tensor.shape[2:], 1)
        fan_in = tensor.shape[1] * kernel_prod
        fan_out = tensor.shape[0] * kernel_prod

    else:
        raise ValueError(
            f"Tensor with dims {tensor.ndim} is not supported. "
            + "Must be at least 2D."
        )

    return fan_in, fan_out


def uniform(tensor: Tensor, a: _Scalar, b: _Scalar) -> None:
    tensor.data = lucid.random.uniform(a, b, tensor.shape).data


def normal(tensor: Tensor, mean: _Scalar, std: _Scalar) -> None:
    tensor.data = lucid.random.randn(tensor.shape).data * std + mean


def constant(tensor: Tensor, val: _Scalar) -> None:
    tensor.data = lucid.ones_like(tensor).data * val


def xavier_uniform(tensor: Tensor, gain: _Scalar) -> None:
    fan_in, fan_out = _calculate_fan_in_and_fan_out(tensor)
    bound = (6 / (fan_in + fan_out)) ** 0.5 * gain
    tensor.data = lucid.random.uniform(-bound, bound, tensor.shape).data


def xavier_normal(tensor: Tensor, gain: _Scalar) -> None:
    fan_in, fan_out = _calculate_fan_in_and_fan_out(tensor)
    std = (2 / (fan_in + fan_out)) ** 0.5 * gain
    tensor.data = lucid.random.randn(tensor.shape).data * std


def kaiming_uniform(tensor: Tensor, mode: str) -> None:
    fan_in, fan_out = _calculate_fan_in_and_fan_out(tensor)
    fan = fan_in if mode == "fan_in" else fan_out
    bound = (6 / fan) ** 0.5
    tensor.data = lucid.random.uniform(-bound, bound, tensor.shape).data


def kaiming_normal(tensor: Tensor, mode: str) -> None:
    fan_in, fan_out = _calculate_fan_in_and_fan_out(tensor)
    fan = fan_in if mode == "fan_in" else fan_out
    std = (2 / fan) ** 0.5
    tensor.data = lucid.random.randn(tensor.shape).data * std



from typing import ClassVar
from lucid import register_model

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor
from lucid.models.objdet.util import nms, DetectionDict


__all__ = ["YOLO_V2", "yolo_v2", "yolo_v2_tiny"]


def _convblock(
    cin: int, cout: int, k: int, s: int = 1, p: int | None = None
) -> list[nn.Module]:
    return [
        nn.Conv2d(
            cin,
            cout,
            kernel_size=k,
            stride=s,
            padding=p if p is not None else "same",
            bias=False,
        ),
        nn.BatchNorm2d(cout),
        nn.LeakyReLU(0.1),
    ]


class _DarkNet_19(nn.Module):
    def __init__(self, num_classes: int = 1000) -> None:
        super().__init__()
        self.num_classes = num_classes

        conv: list[nn.Module] = []
        conv += [
            *_convblock(3, 32, 3),
            nn.MaxPool2d(kernel_size=2, stride=2),
            *_convblock(32, 64, 3),
            nn.MaxPool2d(kernel_size=2, stride=2),
        ]
        conv += [
            *_convblock(64, 128, 3),
            *_convblock(128, 64, 1),
            *_convblock(64, 128, 3),
            nn.MaxPool2d(kernel_size=2, stride=2),
        ]
        conv += [
            *_convblock(128, 256, 3),
            *_convblock(256, 128, 1),
            *_convblock(128, 256, 3),
            nn.MaxPool2d(kernel_size=2, stride=2),
        ]
        conv += [
            *_convblock(256, 512, 3),
            *_convblock(512, 256, 1),
            *_convblock(256, 512, 3),
            *_convblock(512, 256, 1),
            *_convblock(256, 512, 3),
        ]
        self.conv = nn.Sequential(*conv)

        self.classifier = nn.Sequential(
            *_convblock(512, 1024, 3),
            *_convblock(1024, 1024, 3),
            nn.Conv2d(1024, self.num_classes, kernel_size=1),
            nn.AdaptiveAvgPool2d((1, 1)),
            nn.Flatten(),
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(x)
        x = self.classifier(x)
        return x


class _DarkNet_19_Tiny(nn.Module):
    def __init__(self, num_classes: int = 1000):
        super().__init__()
        self.num_classes = num_classes
        self.conv = nn.Sequential(
            *_convblock(3, 16, 3),
            nn.MaxPool2d(kernel_size=2, stride=2),
            *_convblock(16, 32, 3),
            nn.MaxPool2d(kernel_size=2, stride=2),
            *_convblock(32, 64, 3),
            nn.MaxPool2d(kernel_size=2, stride=2),
            *_convblock(64, 128, 3),
            nn.MaxPool2d(kernel_size=2, stride=2),
            *_convblock(128, 256, 3),
            nn.MaxPool2d(kernel_size=2, stride=2),
            *_convblock(256, 512, 3),
            nn.MaxPool2d(kernel_size=2, stride=2),
        )
        self.classifier = nn.Sequential(
            *_convblock(512, 1024, 3),
            nn.Conv2d(1024, self.num_classes, kernel_size=1),
            nn.AdaptiveAvgPool2d((1, 1)),
            nn.Flatten(),
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(x)
        x = self.classifier(x)
        return x


_default_anchors = [
    (1.3221, 1.73145),
    (3.19275, 4.00944),
    (5.05587, 8.09892),
    (9.47112, 4.84053),
    (11.2364, 10.0071),
]


@nn.set_state_dict_pass_attr("darknet_19")
class YOLO_V2(nn.Module):
    default_anchors: ClassVar[list[tuple[float, float]]] = _default_anchors

    def __init__(
        self,
        num_classes: int,
        num_anchors: int = 5,
        anchors: list[tuple[float, float]] | None = None,
        lambda_coord: float = 5.0,
        lambda_noobj: float = 0.5,
        darknet: nn.Module | None = None,
        route_layer: int | None = None,
        image_size: int = 416,
        use_passthrough: bool = True,
    ) -> None:
        super().__init__()
        self.num_classes = num_classes
        self.num_anchors = num_anchors
        self.lambda_coord = lambda_coord
        self.lambda_noobj = lambda_noobj
        self.use_passthrough = use_passthrough

        if anchors is None:
            anchors = YOLO_V2.default_anchors
        self.anchors = nn.Buffer(anchors, dtype=lucid.Float32)

        if darknet is None:
            self.setattr_raw("darknet_19", _DarkNet_19())
            self.darknet = self.darknet_19.conv
            self.route_layer = 28
        else:
            self.darknet = darknet
            if route_layer is None and self.use_passthrough:
                self.route_layer = self._auto_route_index(self.darknet, image_size)
            else:
                self.route_layer = route_layer

        out_c = self._darknet_out_channels(self.darknet)

        if self.use_passthrough:
            route_c = self._layer_out_channels(self.darknet, self.route_layer)
            self.passthrough_conv = nn.Sequential(
                nn.Conv2d(route_c, 64, kernel_size=1),
                nn.BatchNorm2d(64),
                nn.LeakyReLU(0.1),
            )

            self.head_pool = nn.MaxPool2d(kernel_size=2, stride=2)
            self.head_conv1 = nn.Sequential(
                nn.Conv2d(out_c, 1024, kernel_size=3, padding=1),
                nn.BatchNorm2d(1024),
                nn.LeakyReLU(0.1),
            )
            self.head_conv2 = nn.Sequential(
                nn.Conv2d(1024 + 256, 1024, kernel_size=3, padding=1),
                nn.BatchNorm2d(1024),
                nn.LeakyReLU(0.1),
            )

        else:
            self.passthrough_conv = nn.Identity()
            self.head_pool = nn.Identity()

            self.head_conv1 = nn.Sequential(
                nn.Conv2d(out_c, 1024, kernel_size=3, padding=1),
                nn.BatchNorm2d(1024),
                nn.LeakyReLU(0.1),
            )
            self.head_conv2 = nn.Sequential(
                nn.Conv2d(1024, 1024, kernel_size=3, padding=1),
                nn.BatchNorm2d(1024),
                nn.LeakyReLU(0.1),
            )

        self.detector = nn.Conv2d(1024, num_anchors * (5 + num_classes), kernel_size=1)

    @staticmethod
    def _darknet_out_channels(darknet: nn.Module) -> int:
        out_c: int | None = None
        for module in darknet.modules():
            if isinstance(module, nn.Conv2d):
                out_c = module.out_channels
        if out_c is None:
            raise ValueError("No Conv2d layer found in darknet")
        return out_c

    @staticmethod
    def _layer_out_channels(darknet: nn.Module, idx: int) -> int:
        out_c: int | None = None
        for i, module in enumerate(darknet):
            if isinstance(module, nn.Conv2d):
                out_c = module.out_channels
            if i == idx:
                break
        if out_c is None:
            raise ValueError("No Conv2d layer found in darknet")
        return out_c

    @staticmethod
    def _reorg(x: Tensor, stride: int = 2) -> Tensor:
        n, c, h, w = x.shape
        assert h % stride == 0 and w % stride == 0

        x = x.reshape(n, c, h // stride, stride, w // stride, stride)
        x = x.transpose((0, 1, 3, 5, 2, 4))
        return x.reshape(n, c * stride**2, h // stride, w // stride)

    @staticmethod
    def _auto_route_index(darknet: nn.Module, input_hw: int) -> int:
        x = lucid.zeros(1, 3, input_hw, input_hw)
        shapes = []
        for i, layer in enumerate(darknet):
            x = layer(x)
            shapes.append((i, x.shape))

        for i in range(len(shapes) - 1, -1, -1):
            h, w = shapes[i][1][2:]
            if h == 26 and w == 26:
                return i

        return shapes[-1][0]

    def forward(self, x: Tensor) -> Tensor:
        p = None
        for idx, layer in enumerate(self.darknet):
            x = layer(x)
            if self.use_passthrough and idx == self.route_layer:
                p = x

        x = self.head_pool(x)
        x = self.head_conv1(x)

        if self.use_passthrough:
            p = self.passthrough_conv(p)
            p = self._reorg(p)
            x = lucid.concatenate([p, x], axis=1)

        x = self.head_conv2(x)
        x = self.detector(x)

        N, _, H, W = x.shape
        x = x.transpose((0, 2, 3, 1))
        return x.reshape(N, self.num_anchors * (5 + self.num_classes), H, W)

    def get_loss(self, x: Tensor, target: Tensor) -> Tensor:
        N = x.shape[0]
        pred = self.forward(x)

        B, C = self.num_anchors, self.num_classes
        S = pred.shape[2]

        pred = pred.reshape(N, B, 5 + C, S, S).transpose((0, 3, 4, 1, 2))
        target = target.reshape(N, S, S, B, 5 + C)

        obj_mask = target[..., 4:5]
        noobj_mask = 1 - obj_mask

        pred_xy = F.sigmoid(pred[..., 0:2])
        pred_wh = lucid.exp(pred[..., 2:4]) * self.anchors.reshape(1, 1, 1, B, 2)

        pred_obj = F.sigmoid(pred[..., 4:5])
        pred_cls = pred[..., 5:]

        tgt_xy = target[..., 0:2]
        tgt_wh = target[..., 2:4]
        tgt_cls = target[..., 5:]

        loss_xy = F.mse_loss(pred_xy * obj_mask, tgt_xy * obj_mask, reduction="sum")
        loss_wh = F.mse_loss(pred_wh * obj_mask, tgt_wh * obj_mask, reduction="sum")

        loss_obj = F.mse_loss(pred_obj * obj_mask, obj_mask, reduction="sum")
        loss_noobj = F.mse_loss(
            pred_obj * noobj_mask, lucid.zeros_like(pred_obj), reduction="sum"
        )

        cls_loss_vec = F.cross_entropy(
            pred_cls.reshape(-1, C),
            lucid.argmax(tgt_cls, axis=-1).reshape(-1),
            reduction=None,
        )
        loss_cls = (
            cls_loss_vec.reshape(N, S, S, B) * obj_mask.reshape(N, S, S, B)
        ).sum()

        total_loss = (
            self.lambda_coord * (loss_xy + loss_wh)
            + loss_obj
            + self.lambda_noobj * loss_noobj
            + loss_cls
        )
        return total_loss / N

    @lucid.no_grad()
    def predict(
        self, x: Tensor, conf_thresh: float = 0.5, iou_thresh: float = 0.5
    ) -> list[list[DetectionDict]]:
        N, _, H, W = x.shape
        B, C = self.num_anchors, self.num_classes
        stride = self.img_size / H

        x = self.forward(x)
        x = x.reshape(N, B, 5 + C, H, W).transpose((0, 3, 4, 1, 2))

        pred_xy = F.sigmoid(x[..., 0:2])
        pred_wh = lucid.exp(x[..., 2:4]) * self.anchors.reshape(1, 1, 1, B, 2)
        pred_obj = F.sigmoid(x[..., 4])
        pred_cls = F.softmax(x[..., 5:], axis=-1)

        grid_y, grid_x = lucid.meshgrid(lucid.arange(H), lucid.arange(W), indexing="ij")
        grid = lucid.stack([grid_x, grid_y], axis=-1).reshape(1, H, W, 1, 2)

        box_xy = (pred_xy + grid) * stride
        box_wh = pred_wh * stride
        box_xy1 = box_xy - box_wh / 2
        box_xy2 = box_xy + box_wh / 2
        boxes = lucid.concatenate([box_xy1, box_xy2], axis=-1)

        results = []
        for i in range(N):
            boxes_i = boxes[i].reshape(-1, 4)
            scores_i = pred_obj[i].reshape(-1, 1) * pred_cls[i].reshape(-1, C)

            image_preds: list[DetectionDict] = []
            for cl in range(C):
                cls_scores = scores_i[:, cl]
                mask = cls_scores > conf_thresh
                if not mask.any():
                    continue

                cls_boxes = boxes_i[mask]
                cls_scores = cls_scores[mask]

                keep = nms(cls_boxes, cls_scores, iou_thresh)
                for j in keep:
                    image_preds.append(
                        {
                            "box": cls_boxes[j].tolist(),
                            "score": cls_scores[j].item(),
                            "class_id": cl,
                        }
                    )

            results.append(image_preds)

        return results


@register_model
def yolo_v2(num_classes: int = 20, **kwargs) -> YOLO_V2:
    return YOLO_V2(num_classes=num_classes, num_anchors=5, image_size=416, **kwargs)


@register_model
def yolo_v2_tiny(num_classes: int = 20, **kwargs) -> YOLO_V2:
    darknet_tiny = _DarkNet_19_Tiny()
    model = YOLO_V2(
        num_classes=num_classes,
        num_anchors=5,
        image_size=416,
        darknet=darknet_tiny.conv,
        route_layer=None,
        use_passthrough=False,
        **kwargs,
    )
    model.setattr_raw("darknet_19", darknet_tiny)
    return model



import lucid
import lucid.nn.functional

from lucid._tensor import Tensor
from lucid.types import _Scalar, Numeric


def _interpolate_bilinear(
    input_: Tensor, size: tuple[int, int], align_corners: bool = False
) -> Tensor:
    _, _, H, W = input_.shape
    out_h, out_w = size

    scale_h = (H - 1) / (out_h - 1) if align_corners else H / out_h
    scale_w = (W - 1) / (out_w - 1) if align_corners else W / out_w

    indices_h = lucid.arange(out_h).to(input_.device) * scale_h
    indices_w = lucid.arange(out_w).to(input_.device) * scale_w

    if not align_corners:
        indices_h += 0.5 * scale_h
        indices_w += 0.5 * scale_w

    indices_h = indices_h.clip(0, H - 1)
    indices_w = indices_w.clip(0, W - 1)

    top_indices = indices_h.astype(lucid.Int)
    bot_indices = (top_indices + 1).clip(0, H - 1).astype(lucid.Int)
    left_indices = indices_w.astype(lucid.Int)
    right_indices = (left_indices + 1).clip(0, W - 1).astype(lucid.Int)

    h_lerp = indices_h - top_indices
    w_lerp = indices_w - left_indices

    top_left = input_[:, :, top_indices[:, None], left_indices]
    top_right = input_[:, :, top_indices[:, None], right_indices]
    bot_left = input_[:, :, bot_indices[:, None], left_indices]
    bot_right = input_[:, :, bot_indices[:, None], right_indices]

    top = top_left * (1 - w_lerp) + top_right * w_lerp
    bot = bot_left * (1 - w_lerp) + bot_right * w_lerp

    interpolated = top * (1 - h_lerp[:, None]) + bot * h_lerp[:, None]
    return interpolated


def _interpolate_nearest(
    input_: Tensor, size: tuple[int, int], align_corners: bool = False
) -> None:
    _, _, H, W = input_.shape
    device = input_.device
    out_h, out_w = size

    scale_h = H / out_h
    scale_w = W / out_w

    indices_h = (lucid.arange(out_h) * scale_h).clip(0, H - 1).astype(int).to(device)
    indices_w = (lucid.arange(out_w) * scale_w).clip(0, W - 1).astype(int).to(device)

    return input_[:, :, indices_h[:, None], indices_w]


def _interpolate_area(
    input_: Tensor, size: tuple[int, int], align_corners: bool = False
) -> None:
    _, _, H, W = input_.shape
    out_h, out_w = size

    scale_h = H / out_h
    scale_w = W / out_w

    pooled = lucid.nn.functional.avg_pool2d(
        input_,
        kernel_size=(int(scale_h), int(scale_w)),
        stride=(int(scale_h), int(scale_w)),
    )
    return pooled[:, :, out_h, out_w]


def rotate(
    input_: Tensor, angle: float, center: tuple[_Scalar, _Scalar] | None = None
) -> Tensor:
    N, C, H, W = input_.shape

    if center is None:
        center_x = W / 2
        center_y = H / 2
    else:
        center_x, center_y = center

    angle_rad = -angle * (lucid.pi / 180)
    cos_a = lucid.cos(angle_rad).data
    sin_a = lucid.sin(angle_rad).data

    rot_mat = [
        [cos_a, -sin_a, center_x - cos_a * center_x + sin_a * center_y],
        [sin_a, cos_a, center_y - sin_a * center_x - cos_a * center_y],
    ]
    rot_mat = lucid.to_tensor(rot_mat)

    y_coords, x_coords = lucid.arange(H), lucid.arange(W)
    y_grid, x_grid = lucid.meshgrid(y_coords, x_coords, indexing="ij")

    x_flat = x_grid.ravel()
    y_flat = y_grid.ravel()

    ones = lucid.ones_like(x_flat)
    homogen_coords = lucid.stack([x_flat, y_flat, ones])

    new_coords = (rot_mat @ homogen_coords).free()
    new_x = new_coords[0].reshape(H, W)
    new_y = new_coords[1].reshape(H, W)

    new_x = new_x.clip(0, W - 1).astype(lucid.Int)
    new_y = new_y.clip(0, H - 1).astype(lucid.Int)

    rotated_img = lucid.zeros_like(input_, device=input_.device)
    for n in range(N):
        for c in range(C):
            rotated_img[n, c] = input_[n, c, new_y, new_x]

    return rotated_img


def embedding(
    input_: Tensor,
    weight: Tensor,
    padding_idx: int | None = None,
    max_norm: float | None = None,
    norm_type: float = 2.0,
) -> Tensor:
    output = weight[input_.astype(lucid.Int)]
    if padding_idx is not None:
        mask = input_.data == padding_idx
        output *= 1 - mask[..., None]

    if max_norm is not None:
        norm = (output**norm_type).sum(axis=-1, keepdims=True) ** (1 / norm_type)
        scaling = max_norm / (norm + (norm == 0))
        output *= scaling

    return output


def one_hot(
    input_: Tensor, num_classes: int = -1, dtype: Numeric | bool | None = None
) -> Tensor:
    if input_.dtype.base_dtype is not int:
        raise TypeError("one_hot only supports integer input.")
    if num_classes == -1:
        num_classes = lucid.max(input_).item() + 1

    input_flat = input_.reshape(-1)
    N = input_flat.shape[0]

    out_shape = (*input_.shape, num_classes)
    out = lucid.zeros(N, num_classes, device=input_.device, dtype=lucid.Int8)

    arange = lucid.arange(N, device=input_.device, dtype=lucid.Int32)
    out[arange, input_flat.astype(lucid.Int)] = 1

    return (
        out.reshape(out_shape).astype(dtype)
        if dtype is not None
        else out.reshape(out_shape)
    )



from typing import Union
import networkx as nx
import matplotlib.pyplot as plt

import lucid.nn as nn
from lucid._tensor import Tensor


__all__ = ["draw_tensor_graph"]


def draw_tensor_graph(
    tensor: Tensor,
    horizontal: bool = False,
    title: Union[str, None] = None,
    start_id: Union[int, None] = None,
) -> plt.Figure:
    G: nx.DiGraph = nx.DiGraph()
    result_id: int = id(tensor)

    visited: set[int] = set()
    nodes_to_draw: list[Tensor] = []

    def dfs(t: Tensor) -> None:
        if id(t) in visited:
            return
        visited.add(id(t))
        for p in t._prev:
            dfs(p)
        nodes_to_draw.append(t)

    dfs(tensor)

    for t in nodes_to_draw:
        if not t.is_leaf and t._op is not None:
            op_id: int = id(t._op)
            op_label: str = type(t._op).__name__
            G.add_node(op_id, label=op_label, shape="circle", color="lightgreen")
            G.add_edge(op_id, id(t))
            for inp in t._prev:
                G.add_edge(id(inp), op_id)

        shape_label: str = str(t.shape) if t.ndim > 0 else str(t.item())
        if isinstance(t, nn.Parameter):
            color: str = "plum"
        else:
            color = (
                "lightcoral"
                if id(t) == result_id
                else "lightgray" if not t.requires_grad else "lightblue"
            )
        if start_id is not None and id(t) == start_id:
            color = "gold"

        G.add_node(id(t), label=shape_label, shape="rectangle", color=color)

    def grid_layout(
        G: nx.DiGraph, horizontal: bool = False
    ) -> tuple[dict, tuple, float, int]:
        levels: dict[int, int] = {}
        for node in nx.topological_sort(G):
            preds = list(G.predecessors(node))
            levels[node] = 0 if not preds else max(levels[p] for p in preds) + 1

        level_nodes: dict[int, list[int]] = {}
        for node, level in levels.items():
            level_nodes.setdefault(level, []).append(node)

        def autoscale(
            level_nodes: dict[int, list[int]],
            horizontal: bool = False,
            base_size: float = 0.5,
            base_nodesize: int = 500,
        ) -> tuple[tuple[float, float], float, int]:
            num_levels: int = len(level_nodes)
            max_width: int = max(len(nodes) for nodes in level_nodes.values())
            node_count: int = sum(len(nodes) for nodes in level_nodes.values())

            if horizontal:
                fig_w: float = min(32, max(4.0, base_size * num_levels))
                fig_h: float = min(32, max(4.0, base_size * max_width))
            else:
                fig_w = min(32, max(4.0, base_size * max_width))
                fig_h = min(32, max(4.0, base_size * num_levels))

            nodesize: float = (
                base_nodesize
                if node_count <= 100
                else base_nodesize * (100 / node_count)
            )
            fontsize: int = max(5, min(8, int(80 / node_count)))
            return (fig_w, fig_h), nodesize, fontsize

        figsize, nodesize, fontsize = autoscale(level_nodes, horizontal)
        pos: dict[int, tuple[float, float]] = {}
        for level, nodes in level_nodes.items():
            for i, node in enumerate(nodes):
                pos[node] = (
                    (level * 2.5, -i * 2.0) if horizontal else (i * 2.5, -level * 2.0)
                )
        return pos, figsize, nodesize, fontsize

    labels: dict[int, str] = nx.get_node_attributes(G, "label")
    colors: dict[int, str] = nx.get_node_attributes(G, "color")
    shapes: dict[int, str] = nx.get_node_attributes(G, "shape")
    pos, figsize, nodesize, fontsize = grid_layout(G, horizontal)

    fig, ax = plt.subplots(figsize=figsize)

    rect_nodes: list[int] = [n for n in G.nodes() if shapes.get(n) == "rectangle"]
    circ_nodes: list[int] = [n for n in G.nodes() if shapes.get(n) == "circle"]
    rect_colors: list[str] = [colors[n] for n in rect_nodes]

    nx.draw_networkx_nodes(
        G,
        pos,
        nodelist=rect_nodes,
        node_color=rect_colors,
        node_size=nodesize,
        node_shape="s",
        ax=ax,
    )
    nx.draw_networkx_nodes(
        G,
        pos,
        nodelist=circ_nodes,
        node_color="lightgreen",
        node_size=nodesize,
        node_shape="o",
        ax=ax,
    )
    nx.draw_networkx_edges(G, pos, width=0.5, arrows=True, edge_color="gray", ax=ax)
    nx.draw_networkx_labels(G, pos, labels=labels, font_size=fontsize, ax=ax)

    ax.axis("off")
    ax.set_title(title if title is not None else "")

    return fig



import lucid
import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor


__all__ = [
    "DenseNet",
    "densenet_121",
    "densenet_169",
    "densenet_201",
    "densenet_264",
]


class DenseNet(nn.Module):
    def __init__(
        self,
        block_config: tuple[int],
        growth_rate: int = 32,
        num_init_features: int = 64,
        num_classes: int = 1000,
    ) -> None:
        super().__init__()
        self.conv0 = nn.ConvBNReLU2d(
            3, num_init_features, kernel_size=7, stride=2, padding=3, conv_bias=False
        )
        self.pool0 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.blocks = nn.ModuleList()
        self.transitions = nn.ModuleList()

        in_channels = num_init_features
        for i, num_layers in enumerate(block_config):
            block = _DenseBlock(in_channels, num_layers, growth_rate)
            self.blocks.append(block)

            in_channels += num_layers * growth_rate

            if i != len(block_config) - 1:
                out_channels = in_channels // 2
                transition = _TransitionLayer(in_channels, out_channels)
                self.transitions.append(transition)

                in_channels = out_channels

        self.bn_final = nn.BatchNorm2d(in_channels)
        self.relu = nn.ReLU()

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(in_channels, num_classes)

    def forward(self, x: Tensor) -> Tensor:
        x = self.pool0(self.conv0(x))

        for i, block in enumerate(self.blocks):
            x = block(x)
            if i < len(self.transitions):
                x = self.transitions[i](x)

        x = self.avgpool(self.relu(self.bn_final(x)))
        x = x.reshape(x.shape[0], -1)
        x = self.fc(x)

        return x


class _DenseLayer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        growth_rate: int,
        bottleneck: int = 4,
    ) -> None:
        super().__init__()
        self.bn1 = nn.BatchNorm2d(in_channels)
        self.relu1 = nn.ReLU()
        self.conv1 = nn.Conv2d(
            in_channels, bottleneck * growth_rate, kernel_size=1, bias=False
        )

        self.bn2 = nn.BatchNorm2d(bottleneck * growth_rate)
        self.relu2 = nn.ReLU()
        self.conv2 = nn.Conv2d(
            bottleneck * growth_rate,
            growth_rate,
            kernel_size=3,
            padding=1,
            bias=False,
        )

    def forward(self, x: Tensor) -> Tensor:
        out = self.conv1(self.relu1(self.bn1(x)))
        out = self.conv2(self.relu2(self.bn2(out)))

        return lucid.concatenate([x, out], axis=1)


class _DenseBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        num_layers: int,
        growth_rate: int,
        bottleneck: int = 4,
    ) -> None:
        super().__init__()
        layers = [
            _DenseLayer(in_channels + i * growth_rate, growth_rate, bottleneck)
            for i in range(num_layers)
        ]
        self.layers = nn.ModuleList(layers)

    def forward(self, x: Tensor) -> Tensor:
        for layer in self.layers:
            x = layer(x)
        return x


class _TransitionLayer(nn.Module):
    def __init__(self, in_channels: int, out_channels: int) -> None:
        super().__init__()
        self.bn = nn.BatchNorm2d(in_channels)
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)
        self.pool = nn.AvgPool2d(kernel_size=2, stride=2)
        self.relu = nn.ReLU()

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(self.relu(self.bn(x)))
        x = self.pool(x)
        return x


@register_model
def densenet_121(num_classes: int = 1000, **kwargs) -> DenseNet:
    block_config = (6, 12, 24, 16)
    return DenseNet(block_config, num_classes=num_classes, **kwargs)


@register_model
def densenet_169(num_classes: int = 1000, **kwargs) -> DenseNet:
    block_config = (6, 12, 32, 32)
    return DenseNet(block_config, num_classes=num_classes, **kwargs)


@register_model
def densenet_201(num_classes: int = 1000, **kwargs) -> DenseNet:
    block_config = (6, 12, 48, 32)
    return DenseNet(block_config, num_classes=num_classes, **kwargs)


@register_model
def densenet_264(num_classes: int = 1000, **kwargs) -> DenseNet:
    block_config = (6, 12, 64, 48)
    return DenseNet(block_config, num_classes=num_classes, **kwargs)



from enum import Enum
from dataclasses import dataclass
from typing import Any, Dict, Optional

@dataclass(frozen=True)
class WeightEntry:
    url: str
    sha256: str
    tag: str
    dataset: Optional[str] = None
    meta: Optional[Dict[str, Any]] = None

class LeNet_1_Weights(Enum):
    MNIST: WeightEntry
    DEFAULT: WeightEntry

class LeNet_4_Weights(Enum):
    MNIST: WeightEntry
    DEFAULT: WeightEntry

class LeNet_5_Weights(Enum):
    MNIST: WeightEntry
    DEFAULT: WeightEntry

class AlexNet_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class VGGNet_11_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class VGGNet_13_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class VGGNet_16_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class VGGNet_19_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class ResNet_18_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class ResNet_34_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class ResNet_50_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class ResNet_101_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class ResNet_152_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class Wide_ResNet_50_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class Wide_ResNet_101_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class ResNeXt_50_32X4D_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class ResNeXt_101_32X8D_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class ResNeXt_101_64X4D_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class DenseNet_121_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class DenseNet_169_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class DenseNet_201_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class MobileNet_V2_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class MobileNet_V3_Small_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

class MobileNet_V3_Large_Weights(Enum):
    IMAGENET1K: WeightEntry
    DEFAULT: WeightEntry

__all__ = [
    "LeNet_1_Weights",
    "LeNet_4_Weights",
    "LeNet_5_Weights",
    "AlexNet_Weights",
    "VGGNet_11_Weights",
    "VGGNet_13_Weights",
    "VGGNet_16_Weights",
    "VGGNet_19_Weights",
    "ResNet_18_Weights",
    "ResNet_34_Weights",
    "ResNet_50_Weights",
    "ResNet_101_Weights",
    "ResNet_152_Weights",
    "Wide_ResNet_50_Weights",
    "Wide_ResNet_101_Weights",
    "ResNeXt_50_32X4D_Weights",
    "ResNeXt_101_32X8D_Weights",
    "ResNeXt_101_64X4D_Weights",
    "DenseNet_121_Weights",
    "DenseNet_169_Weights",
    "DenseNet_201_Weights",
    "MobileNet_V2_Weights",
    "MobileNet_V3_Small_Weights",
    "MobileNet_V3_Large_Weights",
]



from lucid.types import _DeviceType, _ShapeLike


__all__ = ["UnknownDeviceError", "DeviceMismatchError", "BackwardError"]


class UnknownDeviceError(Exception):
    def __init__(self, device: str) -> None:
        super().__init__(f"Unknown device '{device}'. Must be either 'cpu' or 'gpu'.")


class DeviceMismatchError(Exception):
    def __init__(self, to: _DeviceType, from_: _DeviceType) -> None:
        super().__init__(f"Attempted access of '{to}' tensor from '{from_}' tensor.")


class BackwardError(Exception):
    def __init__(self, shape: _ShapeLike, op: object) -> None:
        super().__init__(
            f"Exception above occurred for tensor of shape {shape} on operation {op}."
        )



import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor


__all__ = ["ZFNet", "zfnet"]


class ZFNet(nn.Module):
    def __init__(self, num_classes: int = 1000) -> None:
        super().__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(3, 96, kernel_size=7, stride=2, padding=1),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(96, 256, kernel_size=5, stride=2, padding=2),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(256, 384, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.Conv2d(384, 384, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.Conv2d(384, 256, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=3, stride=2),
        )

        self.avgpool = nn.AdaptiveAvgPool2d(output_size=(6, 6))
        self.fc = nn.Sequential(
            nn.Dropout(),
            nn.Linear(256 * 6 * 6, 4096),
            nn.ReLU(),
            nn.Dropout(),
            nn.Linear(4096, 4096),
            nn.ReLU(),
            nn.Linear(4096, num_classes),
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(x)
        x = self.avgpool(x)

        x = x.reshape(x.shape[0], -1)
        x = self.fc(x)
        return x


@register_model
def zfnet(num_classes: int = 1000, **kwargs) -> ZFNet:
    return ZFNet(num_classes, **kwargs)



import lucid

from lucid._tensor import Tensor


def linear(input_: Tensor, weight: Tensor, bias: Tensor | None = None) -> Tensor:
    output = input_ @ weight.mT
    if bias is not None:
        output += bias

    return output


def bilinear(
    input_1: Tensor, input_2: Tensor, weight: Tensor, bias: Tensor | None = None
) -> Tensor:
    x1_outer_x2 = input_1[..., :, lucid.newaxis] * input_2[..., lucid.newaxis, :]
    x1_outer_x2_flat = x1_outer_x2.reshape(*x1_outer_x2.shape[:-2], -1)

    weight_flat = weight.reshape(weight.shape[0], -1)
    output = x1_outer_x2_flat @ weight_flat.T

    if bias is not None:
        output += bias

    return output



from typing import Type
import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor


__all__ = ["LeNet", "lenet_1", "lenet_4", "lenet_5"]


class LeNet(nn.Module):
    def __init__(
        self,
        conv_layers: list[dict],
        clf_layers: list[int],
        clf_in_features: int,
        _base_activation: Type[nn.Module] = nn.Tanh,
    ) -> None:
        super().__init__()

        self.conv1 = nn.Sequential(
            nn.Conv2d(1, conv_layers[0]["out_channels"], kernel_size=5),
            _base_activation(),
            nn.AvgPool2d(2, 2),
        )
        self.conv2 = nn.Sequential(
            nn.Conv2d(
                conv_layers[0]["out_channels"],
                conv_layers[1]["out_channels"],
                kernel_size=5,
            ),
            _base_activation(),
            nn.AvgPool2d(2, 2),
        )

        in_features = clf_in_features
        for idx, units in enumerate(clf_layers, start=1):
            self.add_module(f"fc{idx}", nn.Linear(in_features, units))
            self.add_module(f"tanh{idx + 2}", _base_activation())
            in_features = units

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv1(x)
        x = self.conv2(x)
        x = x.reshape(x.shape[0], -1)

        idx = 1
        while hasattr(self, f"fc{idx}"):
            x = getattr(self, f"fc{idx}")(x)
            x = getattr(self, f"tanh{idx + 2}")(x)
            idx += 1

        return x


@register_model
def lenet_1(**kwargs) -> LeNet:
    return LeNet(
        conv_layers=[{"out_channels": 4}, {"out_channels": 12}],
        clf_layers=[10],
        clf_in_features=12 * 4 * 4,
        **kwargs,
    )


@register_model
def lenet_4(**kwargs) -> LeNet:
    return LeNet(
        conv_layers=[{"out_channels": 4}, {"out_channels": 12}],
        clf_layers=[84, 10],
        clf_in_features=12 * 4 * 4,
        **kwargs,
    )


@register_model
def lenet_5(**kwargs) -> LeNet:
    return LeNet(
        conv_layers=[{"out_channels": 6}, {"out_channels": 16}],
        clf_layers=[120, 84, 10],
        clf_in_features=16 * 5 * 5,
        **kwargs,
    )



from typing import Callable
from functools import partial
import math

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid import register_model
from lucid._tensor import Tensor


__all__ = [
    "EfficientFormer",
    "efficientformer_l1",
    "efficientformer_l3",
    "efficientformer_l7",
]


def _to_2_tuple(val: int | float) -> tuple:
    return (val, val)


class _MLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: int | None = None,
        out_features: int | None = None,
        act_layer: type[nn.Module] = nn.GELU,
        norm_layer: type[nn.Module] | None = None,
        bias: bool = True,
        drop: float = 0.0,
    ) -> None:
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        bias = _to_2_tuple(bias)
        drop_probs = _to_2_tuple(drop)

        self.fc1 = nn.Linear(in_features, hidden_features, bias=bias[0])
        self.act = act_layer()
        self.drop1 = nn.Dropout(drop_probs[0])

        self.norm = (
            norm_layer(hidden_features) if norm_layer is not None else nn.Identity()
        )
        self.fc2 = nn.Linear(hidden_features, out_features, bias=bias[1])
        self.drop2 = nn.Dropout(drop_probs[1])

    def forward(self, x: Tensor) -> Tensor:
        x = self.drop1(self.act(self.fc1(x)))
        x = self.drop2(self.fc2(self.norm(x)))
        return x


class _Attention(nn.Module):
    def __init__(
        self,
        dim: int = 384,
        key_dim: int = 32,
        num_heads: int = 8,
        attn_ratio: float = 4.0,
        resolution: int = 7,
    ) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.scale = key_dim**-0.5
        self.key_dim = key_dim
        self.key_attn_dim = key_dim * num_heads
        self.val_dim = int(attn_ratio * key_dim)
        self.val_attn_dim = self.val_dim * num_heads
        self.attn_ratio = attn_ratio

        self.qkv = nn.Linear(dim, self.key_attn_dim * 2 + self.val_attn_dim)
        self.proj = nn.Linear(self.val_attn_dim, dim)

        resolution = _to_2_tuple(resolution)
        y, x = lucid.meshgrid(
            lucid.arange(resolution[0]), lucid.arange(resolution[1]), indexing="ij"
        )
        pos = lucid.stack([y, x]).flatten(axis=1)
        rel_pos = lucid.abs(pos[..., :, None] - pos[..., None, :])
        rel_pos = (rel_pos[0] * resolution[1]) + rel_pos[1]

        self.attention_biases = nn.Parameter(
            lucid.zeros(num_heads, resolution[0] * resolution[1])
        )
        self.attention_bias_idxs: nn.Buffer
        self.register_buffer("attention_bias_idxs", rel_pos.astype(lucid.Int))

    def forward(self, x: Tensor) -> Tensor:
        B, N, _ = x.shape
        qkv = self.qkv(x)
        qkv = qkv.reshape(B, N, self.num_heads, -1).transpose((0, 2, 1, 3))
        q, k, v = qkv.split([self.key_dim, self.key_dim, self.val_dim], axis=3)

        attn = (q @ k.mT) * self.scale
        attn += self.attention_biases[:, self.attention_bias_idxs]
        attn = F.softmax(attn, axis=-1)

        x = (attn @ v).swapaxes(1, 2).reshape(B, N, self.val_attn_dim)
        x = self.proj(x)

        return x


class _Stem(nn.Sequential):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        act_layer: type[nn.Module] = nn.ReLU,
        norm_layer: type[nn.Module] = nn.BatchNorm2d,
    ) -> None:
        super().__init__()
        self.stride = 4

        self.add_module(
            "conv1",
            nn.Conv2d(
                in_channels, out_channels // 2, kernel_size=3, stride=2, padding=1
            ),
        )
        self.add_module("norm1", norm_layer(out_channels // 2))
        self.add_module("act1", act_layer())
        self.add_module(
            "conv2",
            nn.Conv2d(
                out_channels // 2, out_channels, kernel_size=3, stride=2, padding=1
            ),
        )
        self.add_module("norm2", norm_layer(out_channels))
        self.add_module("act2", act_layer())


class _Downsample(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int = 3,
        stride: int = 2,
        padding: int | None = None,
        norm_layer: type[nn.Module] = nn.BatchNorm2d,
    ) -> None:
        super().__init__()
        if padding is None:
            padding = kernel_size // 2
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding)
        self.norm = norm_layer(out_channels)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(x)
        x = self.norm(x)
        return x


class _Flatten(nn.Module):
    def forward(self, x: Tensor) -> Tensor:
        x = x.flatten(axis=2).swapaxes(1, 2)
        return x


class _Pooling(nn.Module):
    def __init__(self, pool_size: int = 3) -> None:
        super().__init__()
        self.pool = nn.AvgPool2d(pool_size, stride=1, padding=pool_size // 2)

    def forward(self, x) -> None:
        return self.pool(x) - x


class _ConvMLPNorm(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: int | None = None,
        out_features: int | None = None,
        act_layer: type[nn.Module] = nn.GELU,
        norm_layer: type[nn.Module] = nn.BatchNorm2d,
        drop: float = 0.0,
    ) -> None:
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features

        self.fc1 = nn.Conv2d(in_features, hidden_features, kernel_size=1)
        self.norm1 = (
            norm_layer(hidden_features) if norm_layer is not None else nn.Identity()
        )
        self.act = act_layer()

        self.fc2 = nn.Conv2d(hidden_features, out_features, kernel_size=1)
        self.norm2 = (
            norm_layer(out_features) if norm_layer is not None else nn.Identity()
        )
        self.drop = nn.Dropout(drop)

    def forward(self, x: Tensor) -> Tensor:
        x = self.act(self.norm1(self.fc1(x)))
        x = self.drop(x)
        x = self.norm2(self.fc2(x))
        x = self.drop(x)

        return x


class _LayerScale(nn.Module):
    def __init__(self, dim: int, init_value: float = 1e-5) -> None:
        super().__init__()
        self.gamma = nn.Parameter(lucid.full(dim, init_value))

    def forward(self, x: Tensor) -> Tensor:
        return x * self.gamma


class _LayerScale2d(_LayerScale):
    def forward(self, x: Tensor) -> Tensor:
        gamma = self.gamma.reshape(1, -1, 1, 1)
        return x * gamma


class _MetaBlock1d(nn.Module):
    def __init__(
        self,
        dim: int,
        mlp_ratio: float = 4.0,
        act_layer: type[nn.Module] = nn.GELU,
        norm_layer: type[nn.Module] = nn.LayerNorm,
        proj_drop: float = 0.0,
        drop_path: float = 0.0,
        layer_scale_init_value: float = 1e-5,
    ) -> None:
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.token_mixer = _Attention(dim)
        self.norm2 = norm_layer(dim)
        self.mlp = _MLP(
            in_features=dim,
            hidden_features=int(dim * mlp_ratio),
            act_layer=act_layer,
            drop=proj_drop,
        )

        self.drop_path = nn.DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.ls1 = _LayerScale(dim, layer_scale_init_value)
        self.ls2 = _LayerScale(dim, layer_scale_init_value)

    def forward(self, x: Tensor) -> Tensor:
        x += self.drop_path(self.ls1(self.token_mixer(self.norm1(x))))
        x += self.drop_path(self.ls2(self.mlp(self.norm2(x))))
        return x


class _MetaBlock2d(nn.Module):
    def __init__(
        self,
        dim: int,
        pool_size: int = 3,
        mlp_ratio: float = 4.0,
        act_layer: type[nn.Module] = nn.GELU,
        norm_layer: type[nn.Module] = nn.BatchNorm2d,
        proj_drop: float = 0.0,
        drop_path: float = 0.0,
        layer_scale_init_value: float = 1e-5,
    ) -> None:
        super().__init__()
        self.token_mixer = _Pooling(pool_size)
        self.ls1 = _LayerScale2d(dim, layer_scale_init_value)
        self.drop_path1 = nn.DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

        self.mlp = _ConvMLPNorm(
            in_features=dim,
            hidden_features=int(dim * mlp_ratio),
            act_layer=act_layer,
            norm_layer=norm_layer,
            drop=proj_drop,
        )
        self.ls2 = _LayerScale2d(dim, layer_scale_init_value)
        self.drop_path2 = nn.DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

    def forward(self, x: Tensor) -> Tensor:
        x += self.drop_path1(self.ls1(self.token_mixer(x)))
        x += self.drop_path2(self.ls2(self.mlp(x)))
        return x


class _EfficientFormerStage(nn.Module):
    def __init__(
        self,
        dim: int,
        dim_out: int,
        depth: int,
        downsample: bool = True,
        num_vit: int = 1,
        pool_size: int = 3,
        mlp_ratio: float = 4.0,
        act_layer: type[nn.Module] = nn.GELU,
        norm_layer: type[nn.Module] = nn.BatchNorm2d,
        norm_layer_cl: type[nn.Module] = nn.LayerNorm,
        proj_drop: float = 0.0,
        drop_path: float = 0.0,
        layer_scale_init_value: float = 1e-5,
    ) -> None:
        super().__init__()
        if downsample:
            self.downsample = _Downsample(dim, dim_out, norm_layer=norm_layer)
            dim = dim_out
        else:
            assert dim == dim_out
            self.downsample = nn.Identity()

        blocks = []
        if num_vit and num_vit >= depth:
            blocks.append(_Flatten())

        for block_idx in range(depth):
            remain_idx = depth - block_idx - 1
            if num_vit and num_vit > remain_idx:
                blocks.append(
                    _MetaBlock1d(
                        dim,
                        mlp_ratio=mlp_ratio,
                        act_layer=act_layer,
                        norm_layer=norm_layer_cl,
                        proj_drop=proj_drop,
                        drop_path=drop_path[block_idx],
                        layer_scale_init_value=layer_scale_init_value,
                    )
                )
            else:
                blocks.append(
                    _MetaBlock2d(
                        dim,
                        pool_size=pool_size,
                        mlp_ratio=mlp_ratio,
                        act_layer=act_layer,
                        norm_layer=norm_layer,
                        proj_drop=proj_drop,
                        drop_path=drop_path[block_idx],
                        layer_scale_init_value=layer_scale_init_value,
                    )
                )
                if num_vit and num_vit == remain_idx:
                    blocks.append(_Flatten())

        self.blocks = nn.Sequential(*blocks)

    def forward(self, x: Tensor) -> Tensor:
        x = self.downsample(x)
        x = self.blocks(x)
        return x


class EfficientFormer(nn.Module):
    def __init__(
        self,
        depths: list[int],
        embed_dims: int | None = None,
        in_channels: int = 3,
        num_classes: int = 1000,
        global_pool: bool = True,
        downsamples: list[bool] | None = None,
        num_vit: int = 0,
        mlp_ratios: float = 4.0,
        pool_size: int = 3,
        layer_scale_init_value: float = 1e-5,
        act_layer: type[nn.Module] = nn.GELU,
        norm_layer: type[nn.Module] = nn.BatchNorm2d,
        norm_layer_cl: type[nn.Module] = nn.LayerNorm,
        drop_rate: float = 0.0,
        proj_drop_rate: float = 0.0,
        drop_path_rate: float = 0.0,
    ) -> None:
        super().__init__()
        self.num_classes = num_classes
        self.global_pool = global_pool

        self.stem = _Stem(in_channels, embed_dims[0], norm_layer=norm_layer)
        prev_dim = embed_dims[0]

        self.num_stages = len(depths)
        last_stage = self.num_stages - 1
        dpr = [
            x.tolist()
            for x in lucid.linspace(0, drop_path_rate, sum(depths)).split(depths)
        ]
        downsamples = downsamples or (False,) + (True,) * (self.num_stages - 1)

        stages = []
        for i in range(self.num_stages):
            stage = _EfficientFormerStage(
                prev_dim,
                embed_dims[i],
                depths[i],
                downsample=downsamples[i],
                num_vit=num_vit if i == last_stage else 0,
                pool_size=pool_size,
                mlp_ratio=mlp_ratios,
                act_layer=act_layer,
                norm_layer_cl=norm_layer_cl,
                norm_layer=norm_layer,
                proj_drop=proj_drop_rate,
                drop_path=dpr[i],
                layer_scale_init_value=layer_scale_init_value,
            )
            prev_dim = embed_dims[i]
            stages.append(stage)

        self.stages = nn.Sequential(*stages)

        self.num_features = embed_dims[-1]
        self.norm = norm_layer_cl(self.num_features)
        self.head_drop = nn.Dropout(drop_rate)
        self.head = (
            nn.Linear(self.num_features, num_classes)
            if num_classes > 0
            else nn.Identity()
        )

        self.apply(self._init_weights)

    def _init_weights(self, m: nn.Module) -> None:
        if isinstance(m, nn.Linear):
            nn.init.normal(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant(m.bias, 0.0)

    def forward_features(self, x: Tensor) -> Tensor:
        x = self.stem(x)
        x = self.stages(x)
        x = self.norm(x)
        return x

    def forward_head(self, x: Tensor, pre_logits: bool = False) -> Tensor:
        if self.global_pool:
            x = x.mean(axis=1)

        x = self.head_drop(x)
        if pre_logits:
            return x

        x = self.head(x)
        return x

    def forward(self, x: Tensor) -> Tensor:
        x = self.forward_features(x)
        x = self.forward_head(x)
        return x


EfficientFormer_width = {
    "l1": (48, 96, 224, 448),
    "l3": (64, 128, 320, 512),
    "l7": (96, 192, 384, 768),
}

EfficientFormer_depth = {
    "l1": (3, 2, 6, 4),
    "l3": (4, 4, 12, 6),
    "l7": (6, 6, 18, 8),
}


@register_model
def efficientformer_l1(num_classes: int = 1000, **kwargs) -> EfficientFormer:
    return EfficientFormer(
        depths=EfficientFormer_depth["l1"],
        embed_dims=EfficientFormer_width["l1"],
        num_vit=1,
        num_classes=num_classes,
        **kwargs,
    )


@register_model
def efficientformer_l3(num_classes: int = 1000, **kwargs) -> EfficientFormer:
    return EfficientFormer(
        depths=EfficientFormer_depth["l3"],
        embed_dims=EfficientFormer_width["l3"],
        num_vit=4,
        num_classes=num_classes,
        **kwargs,
    )


@register_model
def efficientformer_l7(num_classes: int = 1000, **kwargs) -> EfficientFormer:
    return EfficientFormer(
        depths=EfficientFormer_depth["l7"],
        embed_dims=EfficientFormer_width["l7"],
        num_vit=8,
        num_classes=num_classes,
        **kwargs,
    )



from typing import Any, override
from functools import partial

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid import register_model
from lucid._tensor import Tensor


__all__ = [
    "CrossViT",
    "crossvit_tiny",
    "crossvit_small",
    "crossvit_base",
    "crossvit_9",
    "crossvit_15",
    "crossvit_18",
    "crossvit_9_dagger",
    "crossvit_15_dagger",
    "crossvit_18_dagger",
]


def _to_2tuple(val: Any) -> tuple[Any, Any]:
    return (val, val)


def _get_num_patches(img_size: tuple[int, int], patches: tuple[int, int]) -> list[int]:
    return [(i // p) ** 2 for i, p in zip(img_size, patches)]


class _PatchEmbed(nn.Module):
    def __init__(
        self,
        img_size: int = 224,
        patch_size: int = 16,
        in_channels: int = 3,
        embed_dim: int = 768,
        multi_conv: bool = False,
    ) -> None:
        super().__init__()
        img_size = _to_2tuple(img_size)
        patch_size = _to_2tuple(patch_size)
        num_patches = (img_size[1] // patch_size[1]) * (img_size[0] // patch_size[0])

        self.img_size = img_size
        self.patch_size = patch_size
        self.num_patches = num_patches

        if multi_conv:
            if patch_size[0] == 12:
                self.proj = nn.Sequential(
                    nn.Conv2d(in_channels, embed_dim // 4, 7, 4, 3),
                    nn.ReLU(),
                    nn.Conv2d(embed_dim // 4, embed_dim // 2, 3, 3, 0),
                    nn.ReLU(),
                    nn.Conv2d(embed_dim // 2, embed_dim, 3, 1, 1),
                )

            elif patch_size[0] == 16:
                self.proj = nn.Sequential(
                    nn.Conv2d(in_channels, embed_dim // 4, 7, 4, 3),
                    nn.ReLU(),
                    nn.Conv2d(embed_dim // 4, embed_dim // 2, 3, 2, 1),
                    nn.ReLU(),
                    nn.Conv2d(embed_dim // 2, embed_dim, 3, 2, 1),
                )
        else:
            self.proj = nn.Conv2d(
                in_channels, embed_dim, kernel_size=patch_size, stride=patch_size
            )

    def forward(self, x: Tensor) -> Tensor:
        H, W = x.shape[2:]
        if H != self.img_size[0] or W != self.img_size[1]:
            raise ValueError(
                f"Input image size {(H, W)} does not match with {self.img_size}."
            )

        x = self.proj(x).flatten(axis=2).swapaxes(1, 2)
        return x


class _MLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: int | None = None,
        out_features: int | None = None,
        act_layer: type[nn.Module] = nn.GELU,
        norm_layer: type[nn.Module] | None = None,
        bias: bool = True,
        drop: float = 0.0,
        use_conv: bool = False,
    ) -> None:
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features

        bias = _to_2tuple(bias)
        drop_probs = _to_2tuple(drop)
        linear_layer = partial(nn.Conv2d, kernel_size=1) if use_conv else nn.Linear

        self.fc1 = linear_layer(in_features, hidden_features, bias=bias[0])
        self.act = act_layer()
        self.drop1 = nn.Dropout(drop_probs[0])

        self.norm = (
            norm_layer(hidden_features) if norm_layer is not None else nn.Identity()
        )
        self.fc2 = linear_layer(hidden_features, out_features, bias=bias[1])
        self.drop2 = nn.Dropout(drop_probs[1])

    def forward(self, x: Tensor) -> Tensor:
        x = self.drop1(self.act(self.fc1(x)))
        x = self.drop2(self.fc2(self.norm(x)))

        return x


class _LayerScale(nn.Module):
    def __init__(self, dim: int, init_value: float = 1e-5) -> None:
        super().__init__()
        self.gamma = nn.Parameter(init_value * lucid.ones(dim))

    def forward(self, x: Tensor) -> Tensor:
        return x * self.gamma


class _Attention(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int = 8,
        qkv_bias: bool = False,
        qk_norm: bool = False,
        proj_bias: bool = True,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
        norm_layer: type[nn.Module] = nn.LayerNorm,
    ) -> None:
        super().__init__()
        assert dim % num_heads == 0, "dim should be divisible by num_heads."

        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = self.head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.q_norm = norm_layer(self.head_dim) if qk_norm else nn.Identity()
        self.k_norm = norm_layer(self.head_dim) if qk_norm else nn.Identity()

        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim, bias=proj_bias)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: Tensor) -> Tensor:
        B, N, C = x.shape
        qkv = (
            self.qkv(x)
            .reshape(B, N, 3, self.num_heads, self.head_dim)
            .transpose((2, 0, 3, 1, 4))
        )
        q, k, v = qkv.chunk(3)
        q, k = self.q_norm(q), self.k_norm(k)

        attn = q @ k.mT * self.scale
        attn = F.softmax(attn, axis=-1)
        attn = self.attn_drop(attn)

        x = attn @ v
        x = x.reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class _AttentionBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = False,
        qk_norm: bool = False,
        proj_bias: bool = True,
        proj_drop: float = 0.0,
        attn_drop: float = 0.0,
        init_value: float | None = None,
        drop_path: float = 0.0,
        act_layer: type[nn.Module] = nn.GELU,
        norm_layer: type[nn.Module] = nn.LayerNorm,
        mlp_layer: type[nn.Module] = _MLP,
    ) -> None:
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = _Attention(
            dim,
            num_heads,
            qkv_bias,
            qk_norm,
            proj_bias,
            attn_drop,
            proj_drop,
            norm_layer,
        )
        self.ls1 = _LayerScale(dim, init_value) if init_value else nn.Identity()
        self.drop_path1 = nn.DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

        self.norm2 = norm_layer(dim)
        self.mlp = mlp_layer(
            in_features=dim,
            hidden_features=int(dim * mlp_ratio),
            act_layer=act_layer,
            bias=proj_bias,
            drop=proj_drop,
        )
        self.ls2 = _LayerScale(dim, init_value) if init_value else nn.Identity()
        self.drop_path2 = nn.DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

    def forward(self, x: Tensor) -> Tensor:
        x += self.drop_path1(self.ls1(self.attn(self.norm1(x))))
        x += self.drop_path2(self.ls2(self.mlp(self.norm2(x))))

        return x


class _CrossAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int = 8,
        qkv_bias: bool = False,
        qk_scale: float | None = None,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
    ) -> None:
        super().__init__()
        head_dim = dim // num_heads
        self.num_heads = num_heads
        self.scale = qk_scale or head_dim**-0.5

        self.wq = nn.Linear(dim, dim, bias=qkv_bias)
        self.wk = nn.Linear(dim, dim, bias=qkv_bias)
        self.wv = nn.Linear(dim, dim, bias=qkv_bias)

        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: Tensor) -> Tensor:
        B, N, C = x.shape
        q = (
            self.wq(x[:, 0:1, ...])
            .reshape(B, 1, self.num_heads, C // self.num_heads)
            .swapaxes(1, 2)
        )
        k = self.wk(x).reshape(B, N, self.num_heads, C // self.num_heads).swapaxes(1, 2)
        v = self.wv(x).reshape(B, N, self.num_heads, C // self.num_heads).swapaxes(1, 2)

        attn = (q @ k.mT) * self.scale
        attn = F.softmax(attn, axis=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).swapaxes(1, 2).reshape(B, 1, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class _CrossAttentionBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = False,
        qk_scale: float | None = None,
        drop: float = 0.0,
        attn_drop: float = 0.0,
        drop_path: float = 0.0,
        act_layer: type[nn.Module] = nn.GELU,
        norm_layer: type[nn.Module] = nn.LayerNorm,
        has_mlp: bool = True,
    ) -> None:
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = _CrossAttention(
            dim, num_heads, qkv_bias, qk_scale, attn_drop, proj_drop=drop
        )
        self.drop_path = nn.DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

        self.has_mlp = has_mlp
        if has_mlp:
            self.norm2 = norm_layer(dim)
            mlp_hidden_dim = int(dim * mlp_ratio)
            self.mlp = _MLP(dim, mlp_hidden_dim, act_layer=act_layer, drop=drop)

    def forward(self, x: Tensor) -> Tensor:
        x = x[:, 0:1, ...] + self.drop_path(self.attn(self.norm1(x)))
        if self.has_mlp:
            x += self.drop_path(self.mlp(self.norm2(x)))

        return x


class _MultiScaleBlock(nn.Module):
    def __init__(
        self,
        dim: list[int],
        depth: list[int],
        num_heads: list[int],
        mlp_ratio: list[float],
        qkv_bias: bool = False,
        qk_scale: float | None = None,
        drop: float = 0.0,
        attn_drop: float = 0.0,
        drop_path: list[float] | float = 0.0,
        act_layer: type[nn.Module] = nn.GELU,
        norm_layer: type[nn.Module] = nn.LayerNorm,
    ) -> None:
        super().__init__()
        num_branches = len(dim)
        self.num_branches = num_branches

        self.blocks = nn.ModuleList()
        for d in range(num_branches):
            tmp = []
            for i in range(depth[d]):
                tmp.append(
                    _AttentionBlock(
                        dim[d],
                        num_heads[d],
                        mlp_ratio[d],
                        qkv_bias,
                        proj_drop=drop,
                        attn_drop=attn_drop,
                        drop_path=(
                            drop_path[i] if isinstance(drop_path, list) else drop_path
                        ),
                        norm_layer=norm_layer,
                    )
                )
            if len(tmp) != 0:
                self.blocks.append(nn.Sequential(*tmp))

        if len(self.blocks) == 0:
            self.blocks = None

        self.projs = nn.ModuleList()
        for d in range(num_branches):
            if dim[d] == dim[(d + 1) % num_branches]:
                tmp = [nn.Identity()]
            else:
                tmp = [
                    norm_layer(dim[d]),
                    act_layer(),
                    nn.Linear(dim[d], dim[(d + 1) % num_branches]),
                ]
            self.projs.append(nn.Sequential(*tmp))

        self.fusion = nn.ModuleList()
        for d in range(num_branches):
            d_ = (d + 1) % num_branches
            nh = num_heads[d_]

            partial_crossblock = partial(
                _CrossAttentionBlock,
                dim=dim[d_],
                num_heads=nh,
                mlp_ratio=mlp_ratio[d],
                qkv_bias=qkv_bias,
                qk_scale=qk_scale,
                drop=drop,
                attn_drop=attn_drop,
                drop_path=drop_path[i] if isinstance(drop_path, list) else drop_path,
                norm_layer=norm_layer,
            )
            if depth[-1] == 0:
                self.fusion.append(partial_crossblock(has_mlp=False))
            else:
                tmp = []
                for _ in range(depth[-1]):
                    tmp.append(partial_crossblock(has_mlp=False))

                self.fusion.append(nn.Sequential(*tmp))

        self.revert_projs = nn.ModuleList()
        for d in range(num_branches):
            if dim[(d + 1) % num_branches] == dim[d]:
                tmp = [nn.Identity()]
            else:
                tmp = [
                    norm_layer(dim[(d + 1) % num_branches]),
                    act_layer(),
                    nn.Linear(dim[(d + 1) % num_branches], dim[d]),
                ]
            self.revert_projs.append(nn.Sequential(*tmp))

    @override
    def forward(self, xs: list[Tensor]) -> list[Tensor]:
        outs_b = [block(x_) for x_, block in zip(xs, self.blocks)]
        proj_cls_token = [f_proj(x[:, 0:1]) for x, f_proj in zip(outs_b, self.projs)]

        outs = []
        for i in range(self.num_branches):
            tmp = lucid.concatenate(
                [proj_cls_token[i], outs_b[(i + 1) % self.num_branches][:, 1:, ...]],
                axis=1,
            )
            tmp = self.fusion[i](tmp)

            revert_proj_cls_token = self.revert_projs[i](tmp[:, 0:1, ...])
            tmp = lucid.concatenate(
                [revert_proj_cls_token, outs_b[i][:, 1:, ...]], axis=1
            )
            outs.append(tmp)

        return outs


class CrossViT(nn.Module):
    def __init__(
        self,
        img_size: int | list[int] = [224, 224],
        patch_size: list[int] = [12, 16],
        in_channels: int = 3,
        num_classes: int = 1000,
        embed_dim: list[int] = [192, 384],
        depth: list[list[int]] = [[1, 3, 1], [1, 3, 1], [1, 3, 1]],
        num_heads: list[int] = [6, 12],
        mlp_ratio: list[float] = [2.0, 2.0, 4.0],
        qkv_bias: bool = False,
        qk_scale: float | None = None,
        drop_rate: float = 0.0,
        attn_drop_rate: float = 0.0,
        drop_path_rate: float = 0.0,
        norm_layer: type[nn.Module] = nn.LayerNorm,
        multi_conv: bool = False,
    ) -> None:
        super().__init__()
        self.num_classes = num_classes

        if not isinstance(img_size, (list, tuple)):
            img_size = _to_2tuple(img_size)
        self.img_size = img_size

        num_patches = _get_num_patches(img_size, patch_size)
        self.num_branches = len(num_patches)

        self.pos_embed = nn.ParameterList()
        self.cls_token = nn.ParameterList()

        for i in range(self.num_branches):
            self.pos_embed.append(
                nn.Parameter(lucid.zeros(1, 1 + num_patches[i], embed_dim[i]))
            )
            self.cls_token.append(nn.Parameter(lucid.zeros(1, 1, embed_dim[i])))

        self.patch_embed = nn.ModuleList()
        for im_s, p, d in zip(img_size, patch_size, embed_dim):
            self.patch_embed.append(
                _PatchEmbed(im_s, p, in_channels, d, multi_conv=multi_conv)
            )

        self.pos_drop = nn.Dropout(drop_rate)

        total_depth = sum([sum(d[-2:]) for d in depth])
        dpr = [x.item() for x in lucid.linspace(0, drop_path_rate, total_depth)]
        dpr_ptr = 0

        self.blocks = nn.ModuleList()
        for block_cfg in depth:
            cur_depth = max(block_cfg[:-1]) + block_cfg[-1]
            dpr_ = dpr[dpr_ptr : dpr_ptr + cur_depth]

            blk = _MultiScaleBlock(
                dim=embed_dim,
                depth=block_cfg,
                num_heads=num_heads,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                qk_scale=qk_scale,
                drop=drop_rate,
                attn_drop=attn_drop_rate,
                drop_path=dpr_,
                norm_layer=norm_layer,
            )
            dpr_ptr += cur_depth
            self.blocks.append(blk)

        self.norm = nn.ModuleList(
            [norm_layer(embed_dim[i]) for i in range(self.num_branches)]
        )
        self.head = nn.ModuleList()
        for i in range(self.num_branches):
            self.head.append(
                nn.Linear(embed_dim[i], num_classes)
                if num_classes > 0
                else nn.Identity()
            )

        for i in range(self.num_branches):
            if self.pos_embed[i].requires_grad:
                nn.init.normal(self.pos_embed[i], std=0.02)
            nn.init.normal(self.cls_token[i], std=0.02)

        self.apply(self._init_weights)

    def _init_weights(self, m: nn.Module) -> None:
        if isinstance(m, nn.Linear):
            nn.init.normal(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant(m.bias, 0.0)

        elif isinstance(m, nn.LayerNorm):
            nn.init.constant(m.weight, 1.0)
            nn.init.constant(m.bias, 0.0)

    def forward(self, x: Tensor) -> Tensor:
        B, _, H, _ = x.shape
        xs = []
        for i in range(self.num_branches):
            x_ = (
                F.interpolate(x, (self.img_size[i], self.img_size[i]), mode="bilinear")
                if H != self.img_size[i]
                else x
            )
            tmp = self.patch_embed[i](x_)
            cls_tokens = self.cls_token[i].repeat(B, axis=0)

            tmp = lucid.concatenate([cls_tokens, tmp], axis=1)
            tmp += self.pos_embed[i]
            tmp = self.pos_drop(tmp)

            xs.append(tmp)

        for block in self.blocks:
            xs = block(xs)

        xs = [self.norm[i](_x) for i, _x in enumerate(xs)]
        out = [_x[:, 0] for _x in xs]

        logit = [self.head[i](_x) for i, _x in enumerate(out)]
        logit = lucid.stack(logit, axis=0).mean(axis=0)

        return logit


@register_model
def crossvit_tiny(num_classes: int = 1000, **kwargs) -> CrossViT:
    return CrossViT(
        img_size=[240, 224],
        num_classes=num_classes,
        embed_dim=[96, 192],
        depth=[[1, 4, 0] for _ in range(3)],
        num_heads=[3, 3],
        mlp_ratio=[4, 4, 1],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        **kwargs,
    )


@register_model
def crossvit_small(num_classes: int = 1000, **kwargs) -> CrossViT:
    return CrossViT(
        img_size=[240, 224],
        num_classes=num_classes,
        embed_dim=[192, 384],
        depth=[[1, 4, 0] for _ in range(3)],
        num_heads=[6, 6],
        mlp_ratio=[4, 4, 1],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        **kwargs,
    )


@register_model
def crossvit_base(num_classes: int = 1000, **kwargs) -> CrossViT:
    return CrossViT(
        img_size=[240, 224],
        num_classes=num_classes,
        embed_dim=[384, 768],
        depth=[[1, 4, 0] for _ in range(3)],
        num_heads=[12, 12],
        mlp_ratio=[4, 4, 1],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        **kwargs,
    )


@register_model
def crossvit_9(num_classes: int = 1000, **kwargs) -> CrossViT:
    return CrossViT(
        img_size=[240, 224],
        num_classes=num_classes,
        embed_dim=[128, 256],
        depth=[[1, 3, 0] for _ in range(3)],
        num_heads=[4, 4],
        mlp_ratio=[3, 3, 1],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        **kwargs,
    )


@register_model
def crossvit_15(num_classes: int = 1000, **kwargs) -> CrossViT:
    return CrossViT(
        img_size=[240, 224],
        num_classes=num_classes,
        embed_dim=[192, 384],
        depth=[[1, 5, 0] for _ in range(3)],
        num_heads=[6, 6],
        mlp_ratio=[3, 3, 1],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        **kwargs,
    )


@register_model
def crossvit_18(num_classes: int = 1000, **kwargs) -> CrossViT:
    return CrossViT(
        img_size=[240, 224],
        num_classes=num_classes,
        embed_dim=[224, 448],
        depth=[[1, 6, 0] for _ in range(3)],
        num_heads=[7, 7],
        mlp_ratio=[3, 3, 1],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        **kwargs,
    )


@register_model
def crossvit_9_dagger(num_classes: int = 1000, **kwargs) -> CrossViT:
    return CrossViT(
        img_size=[240, 224],
        num_classes=num_classes,
        embed_dim=[128, 256],
        depth=[[1, 3, 0] for _ in range(3)],
        num_heads=[4, 4],
        mlp_ratio=[3, 3, 1],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        multi_conv=True,
        **kwargs,
    )


@register_model
def crossvit_15_dagger(num_classes: int = 1000, **kwargs) -> CrossViT:
    return CrossViT(
        img_size=[240, 224],
        num_classes=num_classes,
        embed_dim=[192, 384],
        depth=[[1, 5, 0] for _ in range(3)],
        num_heads=[6, 6],
        mlp_ratio=[3, 3, 1],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        multi_conv=True,
        **kwargs,
    )


@register_model
def crossvit_18_dagger(num_classes: int = 1000, **kwargs) -> CrossViT:
    return CrossViT(
        img_size=[240, 224],
        num_classes=num_classes,
        embed_dim=[224, 448],
        depth=[[1, 6, 0] for _ in range(3)],
        num_heads=[7, 7],
        mlp_ratio=[3, 3, 1],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        multi_conv=True,
        **kwargs,
    )



from abc import abstractmethod
from pathlib import Path
from typing import Optional, Tuple, Union, ClassVar

from lucid.transforms import Compose
from lucid.data import Dataset
from lucid._tensor import Tensor

import lucid.nn as nn


class DatasetBase(Dataset):
    OPENML_ID: ClassVar[int]

    def __init__(
        self,
        root: Union[str, Path],
        train: Optional[bool] = True,
        download: Optional[bool] = False,
        transform: Optional[nn.Module | Compose] = None,
        target_transform: Optional[nn.Module | Compose] = None,
        test_size: float = 0.2,
        to_tensor: bool = True,
    ) -> None:
        self.root = Path(root)
        self.train = train
        self.transform = transform
        self.target_transform = target_transform
        self.test_size = test_size
        self.to_tensor = to_tensor
        self.root.mkdir(parents=True, exist_ok=True)

        if download:
            self._download()

        if self.train:
            self.data, self.targets = self._load_data("train")
        else:
            self.data, self.targets = self._load_data("test")

    @abstractmethod
    def _download(self) -> None: ...

    @abstractmethod
    def _load_data(self, *args, **kwargs) -> Tuple[Tensor, ...]: ...

    def __len__(self) -> int:
        return len(self.data)



from typing import Literal

from lucid._tensor import Tensor
from lucid.einops import _func

from lucid.types import _EinopsPattern


__all__ = ["rearrange", "reduce", "repeat", "einsum"]


_ReduceStr = Literal["sum", "mean"]


def rearrange(a: Tensor, /, pattern: _EinopsPattern, **shapes: int) -> Tensor:
    return _func.rearrange(pattern, t_shape=a.shape, **shapes)(a)


def reduce(
    a: Tensor, /, pattern: _EinopsPattern, reduction: _ReduceStr = "sum", **shapes: int
) -> Tensor:
    return _func.reduce(pattern, reduction, t_shape=a.shape, **shapes)(a)


def repeat(a: Tensor, /, pattern: _EinopsPattern, **shapes: int) -> Tensor:
    return _func.repeat(pattern, t_shape=a.shape, **shapes)(a)


def einsum(pattern: str, *tensors: Tensor) -> Tensor:
    return _func.einsum(pattern)(*tensors)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="lucid-dl",
    version="2.6.2",
    author="ChanLumerico",
    author_email="greensox284@gmail.com",
    description="Lumerico's Comprehensive Interface for Deep Learning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ChanLumerico/lucid",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.12",
    install_requires=[
        "numpy",
        "pandas",
        "openml",
        "mlx",
    ],
    include_package_data=True,
)



import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor
from lucid.types import _ShapeLike


__all__ = [
    "BatchNorm1d",
    "BatchNorm2d",
    "BatchNorm3d",
    "InstanceNorm1d",
    "InstanceNorm2d",
    "InstanceNorm3d",
    "LayerNorm",
    "GroupNorm",
    "GlobalResponseNorm",
]


class _NormBase(nn.Module):
    def __init__(
        self,
        num_features: int,
        eps: float = 1e-5,
        momentum: float | None = 0.1,
        affine: bool = True,
        track_running_stats: bool = True,
    ) -> None:
        super().__init__()
        self.num_features = num_features
        self.eps = eps
        self.momentum = momentum
        self.affine = affine
        self.track_running_stats = track_running_stats

        if affine:
            weight_ = lucid.ones((num_features,))
            self.weight = nn.Parameter(weight_)

            bias_ = lucid.zeros((num_features,))
            self.bias = nn.Parameter(bias_)
        else:
            self.register_parameter("weight", None)
            self.register_parameter("bias", None)

        if track_running_stats:
            self.running_mean: nn.Buffer
            self.running_var: nn.Buffer

            self.register_buffer("running_mean", lucid.ones((num_features,)))
            self.register_buffer("running_var", lucid.zeros((num_features,)))
        else:
            self.register_buffer("running_mean", None)
            self.register_buffer("running_var", None)

    def reset_running_stats(self) -> None:
        if self.track_running_stats:
            self.running_mean.zero()
            self.running_var.data = lucid.ones((self.num_features,)).data

    def reset_parameters(self) -> None:
        self.reset_running_stats()
        if self.affine:
            self.weight = nn.Parameter(lucid.ones_like(self.weight))
            self.bias = nn.Parameter(lucid.zeros_like(self.bias))

    def extra_repr(self) -> str:
        return (
            f"{self.num_features}, eps={self.eps}, momentum={self.momentum}, "
            f"affine={self.affine}, track_running_stats={self.track_running_stats}"
        )


class _BatchNorm(_NormBase):
    def __init__(
        self,
        num_features: int,
        eps: float = 1e-5,
        momentum: float | None = 0.1,
        affine: bool = True,
        track_running_stats: bool = True,
    ) -> None:
        super().__init__(num_features, eps, momentum, affine, track_running_stats)

    def forward(self, input_: Tensor) -> Tensor:
        return F.batch_norm(
            input_,
            self.running_mean,
            self.running_var,
            self.weight,
            self.bias,
            self.training,
            self.momentum if self.momentum is not None else 0.1,
            self.eps,
        )


class BatchNorm1d(_BatchNorm):
    def forward(self, input_: Tensor) -> Tensor:
        lucid._check_input_dim(input_, dim=3)
        return super().forward(input_)


class BatchNorm2d(_BatchNorm):
    def forward(self, input_: Tensor) -> Tensor:
        lucid._check_input_dim(input_, dim=4)
        return super().forward(input_)


class BatchNorm3d(_BatchNorm):
    def forward(self, input_: Tensor) -> Tensor:
        lucid._check_input_dim(input_, dim=5)
        return super().forward(input_)


class _InstanceNorm(_NormBase):
    def __init__(
        self,
        num_features: int,
        eps: float = 1e-5,
        momentum: float | None = 0.1,
        affine: bool = True,
        track_running_stats: bool = True,
    ) -> None:
        super().__init__(num_features, eps, momentum, affine, track_running_stats)

    def forward(self, input_: Tensor) -> Tensor:
        return F.instance_norm(
            input_,
            self.running_mean,
            self.running_var,
            self.weight,
            self.bias,
            self.training,
            self.momentum if self.momentum is not None else 0.1,
            self.eps,
        )


class InstanceNorm1d(_InstanceNorm):
    def forward(self, input_: Tensor) -> Tensor:
        lucid._check_input_dim(input_, dim=3)
        return super().forward(input_)


class InstanceNorm2d(_InstanceNorm):
    def forward(self, input_: Tensor) -> Tensor:
        lucid._check_input_dim(input_, dim=4)
        return super().forward(input_)


class InstanceNorm3d(_InstanceNorm):
    def forward(self, input_: Tensor) -> Tensor:
        lucid._check_input_dim(input_, dim=5)
        return super().forward(input_)


@nn.auto_repr("normalized_shape", "eps", "elementwise_affine")
class LayerNorm(nn.Module):
    def __init__(
        self,
        normalized_shape: _ShapeLike | int,
        eps: float = 1e-5,
        elementwise_affine: bool = True,
        bias: bool = True,
    ) -> None:
        super().__init__()
        if isinstance(normalized_shape, int):
            normalized_shape = (normalized_shape,)

        self.normalized_shape = tuple(normalized_shape)
        self.eps = eps
        self.elementwise_affine = elementwise_affine

        if elementwise_affine:
            weight_ = lucid.ones(self.normalized_shape)
            self.weight = nn.Parameter(weight_)

            if bias:
                bias_ = lucid.zeros(self.normalized_shape)
                self.bias = nn.Parameter(bias_)
            else:
                self.bias = None
        else:
            self.register_parameter("weight", None)
            self.register_parameter("bias", None)

    def forward(self, input_: Tensor) -> Tensor:
        return F.layer_norm(
            input_, self.normalized_shape, self.weight, self.bias, self.eps
        )


@nn.auto_repr("num_groups", "num_channels", "affine")
class GroupNorm(nn.Module):
    def __init__(
        self, num_groups: int, num_channels: int, eps: float = 1e-5, affine: bool = True
    ) -> None:
        super().__init__()
        self.num_groups = num_groups
        self.num_channels = num_channels
        self.eps = eps
        self.affine = affine

        if self.affine:
            self.weight = nn.Parameter(lucid.ones(num_channels))
            self.bias = nn.Parameter(lucid.zeros(num_channels))
        else:
            self.weight = None
            self.bias = None

    def forward(self, input_: Tensor) -> Tensor:
        return F.group_norm(input_, self.num_groups, self.weight, self.bias, self.eps)


@nn.auto_repr("channels", "eps")
class GlobalResponseNorm(nn.Module):
    def __init__(self, channels: int, eps: float = 1e-6) -> None:
        super().__init__()
        self.gamma = nn.Parameter(lucid.zeros(1, channels, 1, 1))
        self.beta = nn.Parameter(lucid.zeros(1, channels, 1, 1))
        self.eps = eps

    def forward(self, input_: Tensor) -> Tensor:
        return F.global_response_norm(input_, self.gamma, self.beta, self.eps)



from lucid import register_model

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor
from lucid.models.objdet.util import iou


__all__ = ["YOLO_V1", "yolo_v1", "yolo_v1_tiny"]


arch_config = [
    (64, 7, 2, 3),
    "M",
    (192, 3, 1, 1),
    "M",
    (128, 1, 1, 0),
    (256, 3, 1, 1),
    (256, 1, 1, 0),
    (512, 3, 1, 1),
    "M",
    [(256, 1, 1, 0), (512, 3, 1, 1), 4],
    (512, 1, 1, 0),
    (1024, 3, 1, 1),
    "M",
    [(512, 1, 1, 0), (1024, 3, 1, 1), 2],
    (1024, 3, 1, 1),
    (1024, 3, 2, 1),
    (1024, 3, 1, 1),
    (1024, 3, 1, 1),
]

arch_config_tiny = [
    (16, 3, 1, 1),
    "M",
    (32, 3, 1, 1),
    "M",
    (64, 3, 1, 1),
    "M",
    (128, 3, 1, 1),
    "M",
    (256, 3, 1, 1),
    "M",
    (512, 3, 1, 1),
    "M",
    (1024, 3, 1, 1),
    (1024, 3, 1, 1),
    (1024, 3, 1, 1),
]


class _ConvBlock(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, **kwargs) -> None:
        super().__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, bias=False, **kwargs)
        self.bn = nn.BatchNorm2d(out_channels)
        self.act = nn.LeakyReLU(0.1)

    def forward(self, x: Tensor) -> Tensor:
        return self.act(self.bn(self.conv(x)))


class YOLO_V1(nn.Module):
    def __init__(
        self,
        in_channels: int,
        split_size: int,
        num_boxes: int,
        num_classes: int,
        lambda_coord: float = 5.0,
        lambda_noobj: float = 0.5,
        conv_config: list | None = None,
    ) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.split_size = split_size
        self.num_boxes = num_boxes
        self.num_classes = num_classes
        self.lambda_coord = lambda_coord
        self.lambda_noobj = lambda_noobj

        config = conv_config if conv_config is not None else arch_config
        self.darknet = self._create_conv_layers(arch=config)
        self.fcs = nn.Sequential(
            nn.Linear(1024 * split_size**2, 4096),
            nn.LeakyReLU(0.1),
            nn.Linear(4096, split_size**2 * (num_boxes * 5 + num_classes)),
        )

    def _create_conv_layers(self, arch: list) -> nn.Sequential:
        layers = []
        in_channels = self.in_channels
        for cfg in arch:
            if isinstance(cfg, tuple):
                layers.append(
                    _ConvBlock(
                        in_channels,
                        cfg[0],
                        kernel_size=cfg[1],
                        stride=cfg[2],
                        padding=cfg[3],
                    )
                )
                in_channels = cfg[0]

            elif isinstance(cfg, str):
                layers.append(nn.MaxPool2d(kernel_size=2, stride=2))

            elif isinstance(cfg, list):
                conv1, conv2, num_repeats = cfg
                for _ in range(num_repeats):
                    layers.append(
                        _ConvBlock(
                            in_channels,
                            conv1[0],
                            kernel_size=conv1[1],
                            stride=conv1[2],
                            padding=conv1[3],
                        )
                    )
                    layers.append(
                        _ConvBlock(
                            conv1[0],
                            conv2[0],
                            kernel_size=conv2[1],
                            stride=conv2[2],
                            padding=conv2[3],
                        )
                    )
                    in_channels = conv2[0]

        return nn.Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        x = self.darknet(x)
        x = x.flatten(start_axis=1)
        x = self.fcs(x)
        return x

    def get_loss(self, x: Tensor, target: Tensor) -> Tensor:
        N = x.shape[0]
        S, B, C = self.split_size, self.num_boxes, self.num_classes

        pred = self.forward(x).reshape(N, S, S, B * 5 + C)
        target = target.reshape(N, S, S, B * 5 + C)

        obj_mask = target[..., 4:5]
        noobj_mask = 1.0 - obj_mask

        pred_boxes = pred[..., : B * 5].reshape(N, S, S, B, 5)
        target_box = target[..., :5]

        pred_xy = pred_boxes[..., :2]
        pred_wh = pred_boxes[..., 2:4] / 2
        pred_corners = lucid.concatenate(
            [pred_xy - pred_wh, pred_xy + pred_wh], axis=-1
        )

        target_xy = target_box[..., :2]
        target_wh = target_box[..., 2:4] / 2
        target_corners = lucid.concatenate(
            [target_xy - target_wh, target_xy + target_wh], axis=-1
        )

        ious = []
        for b in range(B):
            iou_val = (
                iou(
                    pred_corners[..., b, :].reshape(-1, 4),
                    target_corners.reshape(-1, 4),
                )
                .diagonal()
                .reshape(N, S, S, 1)
            )
            ious.append(iou_val)
        ious = lucid.stack(ious, axis=3)

        best_box = lucid.argmax(ious, axis=3, keepdims=True)
        box_range = lucid.arange(B).reshape(1, 1, 1, B)
        best_mask = (box_range == best_box).astype(lucid.Float32).unsqueeze(axis=-1)

        pred_resp = (pred_boxes * best_mask).sum(axis=3)
        pred_xy = pred_resp[..., :2]
        pred_wh = pred_resp[..., 2:4]
        pred_conf = pred_resp[..., 4:5]

        tgt_xy = target_box[..., :2]
        tgt_wh = target_box[..., 2:4]

        loss_xy = F.mse_loss(pred_xy * obj_mask, tgt_xy * obj_mask, reduction="sum")
        loss_wh = F.mse_loss(
            lucid.sqrt(pred_wh + 1e-6) * obj_mask,
            lucid.sqrt(tgt_wh) * obj_mask,
            reduction="sum",
        )

        pred_conf_all = pred_boxes[..., 4]
        best_ious = lucid.max(ious, axis=3, keepdims=True)

        loss_obj = F.mse_loss(
            pred_conf * obj_mask.squeeze(axis=-1),
            best_ious * obj_mask.squeeze(axis=-1),
            reduction="sum",
        )
        loss_noobj = F.mse_loss(
            pred_conf_all * noobj_mask.squeeze(axis=-1),
            lucid.zeros_like(pred_conf_all),
            reduction="sum",
        )

        pred_cls = pred[..., B * 5 :]
        target_cls = target[..., B * 5 :]
        loss_cls = F.mse_loss(
            pred_cls * obj_mask, target_cls * obj_mask, reduction="sum"
        )

        total_loss = (
            self.lambda_coord * (loss_xy + loss_wh)
            + loss_obj
            + self.lambda_noobj * loss_noobj
            + loss_cls
        )
        return total_loss / N


@register_model
def yolo_v1(num_classes: int = 20, **kwargs) -> YOLO_V1:
    return YOLO_V1(
        in_channels=3,
        split_size=7,
        num_boxes=2,
        num_classes=num_classes,
        lambda_coord=5.0,
        lambda_noobj=0.5,
        **kwargs
    )


@register_model
def yolo_v1_tiny(num_classes: int = 20, **kwargs) -> YOLO_V1:
    return YOLO_V1(
        in_channels=3,
        split_size=7,
        num_boxes=2,
        num_classes=num_classes,
        lambda_coord=5.0,
        lambda_noobj=0.5,
        conv_config=arch_config_tiny,
        **kwargs
    )



import itertools
from typing import Tuple, Optional

import lucid
from lucid._tensor import Tensor


def unfold(
    input_: Tensor,
    filter_size: Tuple[int, ...],
    stride: Tuple[int, ...],
    padding: Tuple[int, ...],
    dilation: Tuple[int, ...],
) -> Tensor:
    input_shape = input_.shape
    if len(input_shape) < 2:
        raise ValueError("Input tensor must have at least 2 dimensions (N and C).")
    N, C, *spatial_dims = input_shape
    D = len(spatial_dims)

    if not (len(filter_size) == len(stride) == len(padding) == len(dilation) == D):
        raise ValueError(
            "filter_size, stride, padding, and dilation must match spatial dims."
        )

    out_dims = []
    for i in range(D):
        eff_k = dilation[i] * (filter_size[i] - 1) + 1
        o = (spatial_dims[i] + 2 * padding[i] - eff_k) // stride[i] + 1
        if o <= 0:
            raise ValueError(f"Non-positive output dim for axis {i}: {o}")
        out_dims.append(o)

    pad_config = [(0, 0), (0, 0)] + [(padding[i], padding[i]) for i in range(D)]
    x = lucid.pad(input_, pad_config)

    offsets = list(itertools.product(*[range(k) for k in filter_size]))
    patches = []
    for off in offsets:
        sl = [slice(None), slice(None)]
        for d in range(D):
            start = off[d] * dilation[d]
            end = start + stride[d] * out_dims[d]
            sl.append(slice(start, end, stride[d]))

        p = x[tuple(sl)]
        p = p.unsqueeze(axis=2)
        patches.append(p)

    col = lucid.concatenate(patches, axis=2)
    new_shape = [N, C] + list(filter_size) + out_dims
    col = col.reshape(new_shape)

    perm = [0] + list(range(2 + D, 2 + 2 * D)) + [1] + list(range(2, 2 + D))
    col = col.transpose(perm)

    N_out = N
    for o in out_dims:
        N_out *= o
    C_filt = C
    for k in filter_size:
        C_filt *= k

    return col.reshape((N_out, C_filt))


def _im2col_conv(
    input_: Tensor,
    weight: Tensor,
    bias: Optional[Tensor],
    stride: Tuple[int, ...],
    padding: Tuple[int, ...],
    dilation: Tuple[int, ...],
    groups: int = 1,
) -> Tensor:
    N, C_in, *input_spatial = input_.shape
    C_out, C_in_div_g, *filter_size = weight.shape
    D = len(filter_size)

    if C_in % groups != 0 or C_out % groups != 0 or C_in_div_g * groups != C_in:
        raise ValueError("Inconsistent channel/group configuration.")

    out_dims = []
    for i in range(D):
        eff = dilation[i] * (filter_size[i] - 1) + 1
        o = (input_spatial[i] + 2 * padding[i] - eff) // stride[i] + 1
        if o <= 0:
            raise ValueError(f"Non-positive output dim for axis {i}: {o}")
        out_dims.append(o)

    col = unfold(input_, filter_size, stride, padding, dilation)

    prod_filter = 1
    for k in filter_size:
        prod_filter *= k

    C_in_g = C_in // groups
    C_out_g = C_out // groups

    weight_rs = weight.reshape(groups, C_out_g, C_in_g * prod_filter)
    N_out = N
    for o in out_dims:
        N_out *= o
    col_rs = col.reshape(N_out, groups, C_in_g * prod_filter)

    outs = []
    for g in range(groups):
        c_g = col_rs[:, g, :]
        w_g = weight_rs[g]
        out = lucid.einops.einsum("nk,ok->no", c_g, w_g)
        outs.append(out)

    out_cat = lucid.concatenate(outs, axis=1)

    new_shape = [N] + out_dims + [C_out]
    out_nd = out_cat.reshape(new_shape)

    perm = [0, D + 1] + list(range(1, 1 + D))
    out_final = out_nd.transpose(perm)

    if bias is not None:
        bias_sh = [1, C_out] + [1] * D
        out_final += bias.reshape(tuple(bias_sh))

    return out_final


def conv(
    input_: Tensor,
    weight: Tensor,
    bias: Optional[Tensor],
    stride: Tuple[int, ...],
    padding: Tuple[int, ...],
    dilation: Tuple[int, ...],
    groups: int,
) -> Tensor:
    if len(input_.shape) < 3 or len(weight.shape) < 3:
        raise ValueError("Input and weight tensors must have at least 3 dimensions.")

    if len(stride) != len(padding) or len(stride) != len(dilation):
        raise ValueError("Stride, padding, and dilation must have the same length.")

    return _im2col_conv(input_, weight, bias, stride, padding, dilation, groups)


def _upsample_nd(input_: Tensor, stride: Tuple[int, ...]) -> Tensor:
    x = input_
    D = len(stride)
    for d in range(D):
        axis = d + 2
        s = stride[d]
        if s <= 1:
            continue

        patches = []
        L = x.shape[axis]
        for i in range(L):
            sl = [slice(None)] * x.ndim
            sl[axis] = slice(i, i + 1)
            patches.append(x[tuple(sl)])

            if i < L - 1:
                zero_shape = list(x.shape)
                zero_shape[axis] = s - 1
                patches.append(lucid.zeros(*zero_shape, device=x.device))

        x = lucid.concatenate(patches, axis=axis)

    return x


def conv_transpose(
    input_: Tensor,
    weight: Tensor,
    bias: Optional[Tensor],
    stride: Tuple[int, ...],
    padding: Tuple[int, ...],
    output_padding: Tuple[int, ...],
    dilation: Tuple[int, ...],
    groups: int = 1,
) -> Tensor:
    C_in = input_.shape[1]
    C_in_w, C_out_g, *kernel_size = weight.shape
    D = len(kernel_size)

    if len(output_padding) != D:
        raise ValueError(
            f"output_padding length must be {D}, got {len(output_padding)}"
        )
    for i, op in enumerate(output_padding):
        if op < 0 or op >= stride[i]:
            raise ValueError(
                f"output_padding[{i}] must be in range [0, {stride[i]}), got {op}"
            )
    if C_in_w != C_in:
        raise ValueError("Weight's first dimension must match input channels.")
    if C_in % groups != 0:
        raise ValueError("Input channels must be divisible by groups.")
    C_in_g = C_in // groups

    pad_ = tuple(dilation[i] * (kernel_size[i] - 1) - padding[i] for i in range(D))

    outputs = []
    for g in range(groups):
        inp_g = input_[:, g * C_in_g : (g + 1) * C_in_g]
        w_seg = weight[g * C_in_g : (g + 1) * C_in_g]

        perm = [1, 0] + list(range(2, 2 + D))
        w_t = w_seg.transpose(perm)
        flip_slices = [slice(None), slice(None)] + [
            slice(None, None, -1) for _ in range(D)
        ]
        w_t = w_t[tuple(flip_slices)]
        ups = _upsample_nd(inp_g, stride)

        if any(op > 0 for op in output_padding):
            for d, op in enumerate(output_padding):
                if op > 0:
                    axis = d + 2
                    zero_shape = list(ups.shape)
                    zero_shape[axis] = op
                    zeros = lucid.zeros(*zero_shape, dtype=ups.dtype, device=ups.device)
                    ups = lucid.concatenate([ups, zeros], axis=axis)

        out_g = _im2col_conv(
            ups,
            w_t,
            bias=None,
            stride=(1,) * D,
            padding=pad_,
            dilation=dilation,
            groups=1,
        )
        outputs.append(out_g)

    output = lucid.concatenate(outputs, axis=1)
    if bias is not None:
        b_shape = [1, C_out_g * groups] + [1] * D
        output = output + bias.reshape(tuple(b_shape))

    return output
